# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""ATBU interface for Google Cloud Storage as used via ResumableUpload and
ChunkedDownload. At the time of ATBU initial implementation, libcloud did not
support GCS multipart upload.
"""
from concurrent.futures import ThreadPoolExecutor
import os
import queue
import threading
import time
from typing import Iterator
from uuid import uuid4
import requests
from google.cloud import storage
from google.oauth2 import service_account
from google.auth.transport.requests import AuthorizedSession
from google.resumable_media.requests import ResumableUpload, ChunkedDownload
from google.resumable_media.common import InvalidResponse, RetryStrategy
import google.api_core.exceptions

from ..credentials import CredentialByteArray
from .base import (
    AUTO_FIND_CREATE_CONTAINER_INDICATOR_CHAR,
    StorageContainerInterface,
    StorageInterface,
    StorageObjectInterface,
    DEFAULT_CHUNK_UPLOAD_SIZE,
)
from ...common.constants import *
from ..storage_interface import base
from ...common.exception import RetryLimitReached, ObjectDoesNotExistError
from ..config import *

DEFAULT_RETRY_EXCEPTIONS = base.DEFAULT_RETRY_EXCEPTIONS + (InvalidResponse,)
TWO_DAYS_IN_SECONDS = 60 * 60 * 24 * 2
DEFAULT_RETRY_SECONDS = TWO_DAYS_IN_SECONDS


class WriteableQueueIterator:
    """Emulates stream.write(bytes) that ChunkedUpload expects.
    Allows to be read as iterator.
    """

    def __init__(self):
        self.queue = queue.Queue()
        self.finished = False

    def write(self, b):
        self.queue.put(b)

    def _checkiter(self):
        if self.finished:
            raise StopIteration(
                f"Iterator already observed 0-byte chunk, iteration already finished."
            )

    def __iter__(self):
        self._checkiter()
        return self

    def __next__(self):
        self._checkiter()
        chunk = self.queue.get()
        if len(chunk) == 0:
            self.finished = True
        return chunk


class GoogleStorageObjectInterface(StorageObjectInterface):
    def __init__(self, storage_object: storage.Blob):
        super().__init__()
        self.storage_object = storage_object

    @property
    def size(self):
        return self.storage_object.size

    @property
    def hash(self):
        return self.storage_object.md5_hash

    @property
    def name(self):
        return self.storage_object.name


class GoogleStorageContainerInterface(StorageContainerInterface):
    def __init__(self, container: storage.Bucket):
        super().__init__()
        self.container = container

    @property
    def name(self) -> str:
        return self.container.name

    def get_object(self, object_name: str):
        try:
            blob = self.container.get_blob(blob_name=object_name)
        except google.api_core.exceptions.RetryError as ex:
            raise RetryLimitReached(
                f"Retry limit reached while trying to access blob {object_name} of container {self.container.name}. {exc_to_string(ex)}",
                ex,
            ) from ex
        if blob is None:
            raise ObjectDoesNotExistError(
                f"The object {object_name} does not exist or was not found."
            )
        object_interface = GoogleStorageObjectInterface(storage_object=blob)

        return object_interface

    def delete_object(self, object_name: str):
        try:
            self.container.delete_blob(object_name)
        except google.cloud.exceptions.NotFound as ex:
            raise ObjectDoesNotExistError(
                f"The object {object_name} does not exist or was not found. {exc_to_string(ex)}"
            ).with_traceback(ex.__traceback__) from ex

    def list_objects(self, prefix: str = None) -> list[StorageObjectInterface]:
        blob_list = self.container.list_blobs(prefix=prefix)
        result: list[StorageObjectInterface] = []
        for b in blob_list:
            result.append(GoogleStorageObjectInterface(storage_object=b))
        return result


class GoogleStorageInterface(StorageInterface):

    class_lock = threading.Lock()
    _thread_exec = None

    @classmethod
    def _class_init(cls):
        with cls.class_lock:
            if cls._thread_exec is None:
                cls._thread_exec = ThreadPoolExecutor(
                    thread_name_prefix="GoogleStorageInterface-downloader"
                )

    def __init__(self, storage_def):
        GoogleStorageInterface._class_init()
        super().__init__()
        self.storage_def: dict = storage_def
        self.driver_config: dict = self.storage_def[CONFIG_SECTION_DRIVER]
        self.project_name = self.driver_config[CONFIG_VALUE_NAME_DRIVER_STORAGE_PROJECT]
        self.scoped_credentials = self._create_credential()

    def _create_credential(self):
        password_type = self.driver_config.get(CONFIG_PASSWORD_TYPE)
        if password_type is None:
            raise f"Credential type not found in configuration."
        secret = self.driver_config[CONFIG_VALUE_NAME_DRIVER_STORAGE_SECRET]
        if isinstance(secret, (CredentialByteArray, bytearray, bytes)):
            secret = secret.decode("utf-8")
        if password_type == CONFIG_PASSWORD_TYPE_FILENAME:
            # For OAuth 2 .json, referencing the file location would be more likely.
            credentials = service_account.Credentials.from_service_account_file(
                filename=secret
            )
        elif password_type == CONFIG_PASSWORD_TYPE_ENVVAR:
            if secret is None or secret == "":
                raise ValueError(f"Environment variable not found")
            cred_filename = os.getenv(secret)
            credentials = service_account.Credentials.from_service_account_file(
                cred_filename
            )
        else:
            raise ValueError(f"Unknown credential type '{password_type}'.")
        scoped_credentials = credentials.with_scopes(
            ["https://www.googleapis.com/auth/devstorage.read_write"]
        )
        return scoped_credentials

    def _create_session(self):
        return AuthorizedSession(credentials=self.scoped_credentials)

    def get_container(self, container_name: str) -> StorageContainerInterface:
        storage_client = storage.Client(
            project=self.project_name, credentials=self.scoped_credentials
        )
        bucket = storage_client.bucket(container_name)
        return GoogleStorageContainerInterface(container=bucket)

    def _create_container(self, container_name: str) -> StorageContainerInterface:
        try:
            storage_client = storage.Client(
                project=self.project_name, credentials=self.scoped_credentials
            )
            bucket = storage_client.create_bucket(bucket_or_name=container_name)
            return GoogleStorageContainerInterface(container=bucket)
        except google.api_core.exceptions.Conflict as ex:
            raise InvalidContainerNameError(
                f"Cannot create container '{container_name}' due to a conflict: {exc_to_string(ex)}"
            ) from ex

    def create_container(self, container_name: str) -> StorageContainerInterface:
        if container_name[-1] != AUTO_FIND_CREATE_CONTAINER_INDICATOR_CHAR:
            return self._create_container(container_name=container_name)
        else:
            base_name = container_name[:-1]
            max_attempts = 100
            attempts_remaining = max_attempts
            while True:
                attempts_remaining -= 1
                candidate = f"{base_name}-{str(uuid4())}"
                try:
                    ci = self._create_container(container_name=candidate)
                    return ci
                except (ContainerAlreadyExistsError, InvalidContainerNameError) as ex:
                    if attempts_remaining <= 0:
                        raise ContainerAutoCreationFailed(
                            f"Automatic container creation failed after {max_attempts} attempts. "
                            f"{exc_to_string(ex)}"
                        ) from ex

    def delete_container(self, container_name: str):
        try:
            storage_client = storage.Client(
                project=self.project_name, credentials=self.scoped_credentials
            )
            b = storage_client.get_bucket(container_name)
            b.delete()
        except Exception:  # pylint: disable=try-except-raise
            raise  # TBD any customizations.

    def upload_stream_to_object(
        self,
        container: StorageContainerInterface,
        object_name: str,
        stream,
        source_path: str,
    ):
        session = self._create_session()
        session.headers["x-goog-project-id"] = self.project_name
        url_template = (
            "https://www.googleapis.com/upload/storage/v1/b/{bucket}/o?"
            "uploadType=resumable"
        )
        upload_url = url_template.format(
            bucket=self.storage_def[CONFIG_VALUE_NAME_CONTAINER]
        )
        metadata = {"name": object_name}
        upload = ResumableUpload(upload_url, DEFAULT_CHUNK_UPLOAD_SIZE)
        upload._retry_strategy = RetryStrategy(  # pylint: disable=protected-access
            max_sleep=30, max_cumulative_retry=DEFAULT_RETRY_SECONDS
        )
        response = upload.initiate(
            transport=session,
            stream=stream,
            metadata=metadata,
            content_type="application/octet-stream",
            stream_final=False,
        )
        logging.debug(
            f"Uploading: project={self.project_name} container={container.name} "
            f"object={object_name}: UploadID={response.headers['X-GUploader-UploadID']} "
            f"from source_path={source_path}"
        )
        logging.info(f"Backing up: {source_path}")
        perf_seconds_start = time.perf_counter()
        chunk_number = 1
        pos_before_xmit = 0
        while not upload.finished:
            last_bytes_uploaded = upload.bytes_uploaded
            try:
                seconds_running = time.perf_counter() - perf_seconds_start
                logging.debug(
                    f"Uploading chunk {chunk_number} at {seconds_running:.3f} seconds into the backup. "
                    f"source_path={source_path} dest={container.name}:{object_name}"
                )
                pos_before_xmit = stream.tell()
                response = upload.transmit_next_chunk(session)
                logging.debug(
                    f"upload_stream_to_object: chunk upload response={response}"
                )
                logging.debug(
                    f"upload_stream_to_object: chunk bytes uploaded={upload.bytes_uploaded - last_bytes_uploaded}"
                )
                logging.debug(
                    f"upload_stream_to_object: total bytes uploaded={upload.bytes_uploaded}"
                )
                chunk_number += 1
            except Exception as ex:
                # We cannot be certain of the exception being caused by final timeout or not.
                # Generally, we may observe MaxRetryError but that is for the inner connection
                # error so it is a MaxRetryError of each attempt, where we will see the last one.
                # There is currently no "RetryStrategyLimitReached" exception, for example, so we
                # sample time running to determine if the retry strategy limit has been reached.
                seconds_running = time.perf_counter() - perf_seconds_start
                pos_after_xmit = stream.tell()
                chunk_size = pos_after_xmit - pos_before_xmit
                logging.error(
                    f"Chunk upload of {chunk_size} bytes via transmit_next_chunk failed. seconds_running={seconds_running:.3f}"
                    f"upload.invalid={upload.invalid} pos_before={pos_before_xmit} pos_after={pos_after_xmit} {exc_to_string(ex)}"
                )
                # Has retry strategy limit been reached?
                if seconds_running >= DEFAULT_RETRY_SECONDS:
                    raise RetryLimitReached(
                        f"Backup file write failed after {seconds_running:.3f} overall retry time. {exc_to_string(ex)}",
                        ex,
                    ) from ex
                # Attempt to recover.
                if not upload.invalid:
                    # When transmit_next_chunk exhausts timeout, it can raise without
                    # setting invalid=True which will cause recover to fail. Alteratively,
                    # we could do the following and skip recover which works...
                    #     if stream.tell() != pos_before_xmit:
                    #         stream.seek(pos_before_xmit, SEEK_SET)
                    # The following _make_invalid is what ResumableUpload does internally
                    # for other exceptions so is done here.
                    upload._make_invalid()  # pylint: disable=protected-access
                # Attempt to recover. This seems to essentially amount to pinging the server
                # for its notion of the current range for the upload, where that by itself
                # validates connection is good, and the mutual understanding, and recover
                # rewinds stream back to start of last read chunk so transmit_next_chunk
                # can perform the re-read and retry.
                response: requests.Response = upload.recover(session)
                logging.warning(f"recover result={response}")
        # json_response = response.json()
        # logging.debug(f"json_response={json_response}")

    def download_object_as_stream(
        self, storage_object: GoogleStorageObjectInterface, chunk_size: int
    ) -> Iterator[bytes]:
        def _internal_feed_stream_download(download, stream):
            perf_seconds_start = time.perf_counter()
            try:
                while not download.finished:
                    try:
                        response = download.consume_next_chunk(
                            session
                        )  # pylint: disable=unused-variable
                        # logging.debug(f"finished={download.finished} read={len(r.content)} chunk_size={chunk_size}")
                        # logging.debug(f"Content-Length={r.headers['Content-Length']}")
                        # logging.debug(f"Content-Range={r.headers['Content-Range']}")
                        # logging.debug(f"status_code: r={r.status_code}")
                    except Exception as ex:
                        seconds_running = time.perf_counter() - perf_seconds_start
                        if seconds_running >= DEFAULT_RETRY_SECONDS:
                            raise RetryLimitReached(
                                f"Backup file write failed after {seconds_running:.3f} overall retry time. {exc_to_string(ex)}",
                                ex,
                            ) from ex
                        if download.invalid:
                            raise RetryLimitReached(
                                f"The download object is invalid, cannot retry. Backup file write failed after {seconds_running:.3f} overall retry time. ex={ex}",
                                ex,
                            ) from ex
            finally:
                if not stream.finished:
                    stream.write(bytes())

        session = self._create_session()
        session.headers["x-goog-project-id"] = self.project_name
        url_template = "https://www.googleapis.com/download/storage/v1/b/{bucket}/o/{blob_name}?alt=media"
        download_url = url_template.format(
            bucket=self.storage_def[CONFIG_VALUE_NAME_CONTAINER],
            blob_name=storage_object.name,
        )
        stream = WriteableQueueIterator()
        download = ChunkedDownload(
            media_url=download_url, chunk_size=chunk_size, stream=stream
        )
        download._retry_strategy = RetryStrategy(  # pylint: disable=protected-access
            max_sleep=30, max_cumulative_retry=DEFAULT_RETRY_SECONDS
        )

        future = GoogleStorageInterface._thread_exec.submit(
            _internal_feed_stream_download, download, stream
        )
        try:
            for c in stream:
                if not future.running():
                    future.result()
                yield c
        finally:
            future.result()

    @property
    def retry_exceptions(self) -> tuple:
        return self._retry_exceptions



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Persistent file information database-related functionality allows
gathering of persistent file information, placing it into a single
json "database" file.

There was a second revision of persistent info functionality incorporated
into ATBU so you may see redundancy or superfluous constructs within.
"""
# pylint: disable=missing-class-docstring
from dataclasses import dataclass
import os
import json
import logging
from collections import defaultdict

from ..common.exception import *
from ..common.constants import *
from ..common.file_info import (
    FileInformationPersistent,
    get_loc_persistent_file_info,
    LocationFileInfoUpdater,
)
from ..common.hasher import GlobalHasherDefinitions, HasherDefinitions


class CustomEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, FileInformationPersistent):
            return o.to_serialization_dict()
        return json.JSONEncoder.default(self, o)


class CustomDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, obj):  # pylint: disable=method-hidden,no-self-use
        if "_type" not in obj:
            return obj
        object_type = obj["_type"]
        if object_type == "FileInformationPersistent":
            return FileInformationPersistent(
                path=obj["path"],
                info_current=obj["info_current"],
                info_history=obj["info_history"],
                is_loaded_from_db=True,
            )
        elif (
            object_type == "FileInformation"
        ):  # FileInformationPersistent used to be FileInformation (when used w/.atbu files)
            return FileInformationPersistent(
                path=obj["path"],
                info_current=obj["info_current"],
                info_history=obj["info_history"],
                is_loaded_from_db=True,
            )
        return obj


def get_location_info_from_database(
    database_json_path,
) -> defaultdict[str, list[FileInformationPersistent]]:
    with open(database_json_path, "r", encoding="utf-8") as database_file:
        return json.load(database_file, cls=CustomDecoder)


def file_info_list_to_digest_dict(
    file_info_list: list[FileInformationPersistent],
):
    digest_dict: defaultdict[str, list[FileInformationPersistent]] = defaultdict(
        list[FileInformationPersistent]
    )
    for file_info in file_info_list:
        digest_dict[file_info.primary_digest].append(file_info)
    return digest_dict


def file_info_list_to_ncpath_dict(
    file_info_list: list[FileInformationPersistent],
) -> dict[FileInformationPersistent]:
    ncpath_dict: defaultdict[str, FileInformationPersistent] = {}
    for file_info in file_info_list:
        fi = ncpath_dict.get(file_info.nc_path)
        if fi:
            raise InvalidStateError(
                f"Expecting only one file info per path: path={fi.path}"
            )
        ncpath_dict[file_info.nc_path] = file_info
    return ncpath_dict


def save_database_from_dict_file_info(
    dict_digest_to_file_info_list: defaultdict[str, list[FileInformationPersistent]],
    database_path,
):
    with open(database_path, "w", encoding="utf-8") as database_file:
        database_file.write(
            json.dumps(dict_digest_to_file_info_list, cls=CustomEncoder, indent=4)
        )


def save_database_from_file_info_list(
    file_info_list: list[FileInformationPersistent],
    database_path,
    hasher_defs: HasherDefinitions = None,
):
    if not hasher_defs:
        hasher_defs = GlobalHasherDefinitions()
    digest_dict = file_info_list_to_digest_dict(
        file_info_list=file_info_list,
    )
    save_database_from_dict_file_info(
        dict_digest_to_file_info_list=digest_dict, database_path=database_path
    )


class FileInformationDatabase:
    def __init__(self, source_path: str, persist_type: ATBU_PERSIST_TYPE_HINT) -> None:
        if not os.path.exists(source_path):
            raise InvalidFunctionArgument(
                f"Source path must a be directory but got '{source_path}'."
            )
        self.source_path = os.path.abspath(source_path)
        # If is_source_path_explicit_json_db_file==True means .json is user-created
        # and could pertain to one or more directories not currently online.
        self.is_source_path_explicit_json_db_file = os.path.isfile(self.source_path)
        self.persist_type = persist_type
        if self.is_source_path_explicit_json_db_file:
            self.db_filename = self.source_path
        else:
            self.db_filename = os.path.join(
                self.source_path, ATBU_DEFAULT_PERSISTENT_DB_FILENAME
            )
        self._file_info_list: list[FileInformationPersistent] = None
        self.updater: LocationFileInfoUpdater = None

    def get_file_info(self) -> list[FileInformationPersistent]:
        if self._file_info_list is None:
            raise InvalidStateError(
                f"_file_info_list is None. You must first either load or update the DB."
            )
        return self._file_info_list

    def get_dict_digest_to_fi(self):
        return file_info_list_to_digest_dict(file_info_list=self.get_file_info())

    def get_dict_ncpath_to_fi(self):
        return file_info_list_to_ncpath_dict(file_info_list=self.get_file_info())

    def _get_source_path_file_info_list(self):
        return get_loc_persistent_file_info(locations=self.source_path)

    def _load_per_file_info(self) -> None:
        if self.persist_type != ATBU_PERSIST_TYPE_PER_FILE:
            raise InvalidFunctionArgument(
                f"Expected persist_type={ATBU_PERSIST_TYPE_PER_FILE} but got {self.persist_type}"
            )
        self._file_info_list = self._get_source_path_file_info_list()

    def _load_per_dir_info(self) -> None:
        """Load the .json DB file. Two types of .json DB files based on
        how they were created/used. An user-created DB is created with
        save-db and is not tied to a specific directory, can contain
        files from one or more directories (up to user). The other
        type of DB is that created by update-digests for per-dir cases,
        where it maintains a DB of files within that directory structure.
        This method will load either type and the result is a flat list
        of file information in self._file_info_list.
        """
        if self.persist_type != ATBU_PERSIST_TYPE_PER_DIR:
            raise InvalidStateError(
                f"Expected persist_type={ATBU_PERSIST_TYPE_PER_DIR} but got {self.persist_type}"
            )

        if self.db_filename is None:
            raise InvalidStateError(
                f"db_filename is None, expected db_filename to be a .json db filename."
            )

        # Load file information from the DB file.
        # The info from DB contains path, digests,
        # modified date/time, size, etc.
        digest_to_fi_list = {}
        if os.path.exists(self.db_filename):
            digest_to_fi_list = get_location_info_from_database(
                database_json_path=self.db_filename
            )
        # Flatten to list of file info.
        self._file_info_list = [fi for l in digest_to_fi_list.values() for fi in l]

        # Disallow refresh of file information that comes from
        # an explicitly specified .json DB. This will cause
        # an exception if there's a code logic breaking the
        # assumption we do not expect files to be present
        # with an explicitly specified database.
        if self.is_source_path_explicit_json_db_file:
            for fi in self._file_info_list:
                fi.implicit_refresh_allowed = False

        # If user specifies explicit .json db as a location,
        # the user expects the db to be used as is with the need
        # for files to be present (user captured db earlier, perhaps,
        # before a drive was disconnected). Therefore, skip the
        # following which assumes files exist on the system.
        if not self.is_source_path_explicit_json_db_file:
            # Remove file info for files that no longer exist.
            # Or, in the case of a moved folder, remove info
            # pertaining to the old location.
            nc_source_path = os.path.normcase(self.source_path)
            self._file_info_list = list(
                filter(
                    lambda fi: os.path.isfile(fi.path)
                    and fi.nc_path.startswith(nc_source_path),
                    self._file_info_list,
                )
            )

            #
            # self._file_info_list is now a file info list of files
            # that both exist and are under self.source_path.
            #

            # Index self._file_info_list by path which will be used
            # to find out what file information is missing.
            db_path_to_file_info: dict = {fi.nc_path: fi for fi in self._file_info_list}

            # Get current file system file info. This only gets
            # a bare file info without digest/stat (because they
            # are not updated yet as DB is before being saved).
            fs_file_info_list = self._get_source_path_file_info_list()

            # Index new paths into the self._file_info_list DB info.
            for fi in fs_file_info_list:
                # If not in DB, add to DB.
                if not db_path_to_file_info.get(fi.nc_path):
                    db_path_to_file_info[fi.nc_path] = fi

            self._file_info_list = list(db_path_to_file_info.values())

    def load(self):
        """This loads existing data. TBD relationship with _get_source_path_file_info
        et al which scan files for both persist_types (in prep of updates etc.)
        """
        if self.persist_type == ATBU_PERSIST_TYPE_PER_FILE:
            self._load_per_file_info()
        elif self.persist_type == ATBU_PERSIST_TYPE_PER_DIR:
            self._load_per_dir_info()
        else:
            raise InvalidFunctionArgument(
                f"Unexpected persist_type={self.persist_type}"
            )

    def update(
        self,
        change_detection_type: str,
        update_stale: bool,
        whatif: bool,
    ) -> LocationFileInfoUpdater:
        self.load()
        # Show progress in sorted file path order.
        self._file_info_list.sort(key=lambda fi: fi.nc_path)
        self.updater = LocationFileInfoUpdater(
            file_info_list=self._file_info_list,
            update_stale=update_stale,
            change_detection_type=change_detection_type,
            per_file_persistence=not self.persist_type == ATBU_PERSIST_TYPE_PER_DIR,
            whatif=whatif,
        )
        # Do not update for explicitly specified .json
        # file DBs because files are not necessarily online.
        if not self.is_source_path_explicit_json_db_file:
            self.updater.update()
        return self.updater

    def update_from(self, other_db):
        self._file_info_list = list(other_db.get_file_info())
        # If this instance is a 'per-dir' DB, save the DB.
        # There is no saving to perform for 'per-file' since
        # per-file persistence updates as part of scanning/updating
        # the directories, likely before this update_from is called.
        if self.persist_type == ATBU_PERSIST_TYPE_PER_DIR:
            self.save_db()

    def save_db(self):
        save_database_from_file_info_list(
            database_path=self.db_filename,
            file_info_list=self._file_info_list,
        )


class FileInformationDatabaseCollection:
    def __init__(self, source_path: str, persist_types: list[str]) -> None:
        if not os.path.exists(source_path):
            raise InvalidFunctionArgument(
                f"Source path must a be directory but got '{source_path}'."
            )
        self.source_path = os.path.abspath(source_path)
        self.databases = {}
        for pt in persist_types:
            if pt not in ATBU_PERSIST_TYPES:
                raise InvalidFunctionArgument(
                    f"Expected a valid persist type '{ATBU_PERSIST_TYPES}' but got '{pt}'."
                )
            if self.databases.get(pt):
                raise InvalidFunctionArgument(
                    f"Only one of each persist type is allowed but got more than one of '{pt}'."
                )
            self.databases[pt] = FileInformationDatabase(
                persist_type=pt, source_path=self.source_path
            )

    @property
    def has_per_file_persistence(self):
        for pt in self.databases:
            if pt == ATBU_PERSIST_TYPE_PER_FILE:
                return True
        return False

    def get_combined_db_file_info(self):
        combined_info: dict[str, FileInformationPersistent] = {}
        db: FileInformationDatabase
        for k in [ATBU_PERSIST_TYPE_PER_FILE, ATBU_PERSIST_TYPE_PER_DIR]:
            db = self.databases.get(k)
            if not db:
                continue
            fi_list = db.get_file_info()
            for fi in fi_list:
                if not combined_info.get(fi.nc_path):
                    combined_info[fi.nc_path] = fi
        return combined_info.values()

    def get_dict_nc_path_to_fi(self):
        # pylint: disable=unused-variable
        (
            primary_update_db,
            other_db,
        ) = self._get_primary_other_db()  # pylint: disable=unused-variable
        return primary_update_db.get_dict_ncpath_to_fi()

    def get_dict_digest_to_fi(self):
        # pylint: disable=unused-variable
        (
            primary_update_db,
            other_db,
        ) = self._get_primary_other_db()  # pylint: disable=unused-variable
        return primary_update_db.get_dict_digest_to_fi()

    def _get_primary_other_db(self):
        primary_update_db: FileInformationDatabase = self.databases.get(
            ATBU_PERSIST_TYPE_PER_FILE
        )
        other_db: FileInformationDatabase = self.databases.get(
            ATBU_PERSIST_TYPE_PER_DIR
        )
        if not primary_update_db:
            primary_update_db = self.databases.get(ATBU_PERSIST_TYPE_PER_DIR)
            other_db = None
        if primary_update_db is None:
            raise InvalidFunctionArgument(
                f"Cannot determine the primary update db. DBs={self.databases.keys()}"
            )
        return primary_update_db, other_db

    def load(self):
        primary_update_db, other_db = self._get_primary_other_db()
        if primary_update_db is not None:
            primary_update_db.load()
        if other_db is not None:
            other_db.load()

    def update(
        self,
        change_detection_type: str,
        update_stale: bool,
        whatif: bool,
    ) -> LocationFileInfoUpdater:
        """Update the databases if needed. Returns the updater used for
        access to any stats.
        """
        primary_update_db, other_db = self._get_primary_other_db()
        updater = primary_update_db.update(
            change_detection_type=change_detection_type,
            update_stale=update_stale,
            whatif=whatif,
        )
        if other_db is None:
            # There is no other db, just primary. If primary is per-file it
            # is implicitly already persisted given the update above. If it
            # is a per-dir, it must be saved after the update.
            # Save the per-dir db.
            if (
                primary_update_db.persist_type == ATBU_PERSIST_TYPE_PER_DIR
                and not primary_update_db.is_source_path_explicit_json_db_file
            ):
                primary_update_db.save_db()
        else:
            # Tell other_db to update itself from
            # primary_update_db and save itself to db.
            other_db.update_from(other_db=primary_update_db)
        return updater


def extract_location_info(
    arguments: list[str], min_required: int, max_allowed: int
) -> list[tuple[str, str]]:
    """Helper function to process location arguments captured by argparse."""
    per_arg_to_persist_types = {
        "per-file:": [ATBU_PERSIST_TYPE_PER_FILE],
        "pf:": [ATBU_PERSIST_TYPE_PER_FILE],
        "per-dir:": [ATBU_PERSIST_TYPE_PER_DIR],
        "pd:": [ATBU_PERSIST_TYPE_PER_DIR],
        "per-both:": [ATBU_PERSIST_TYPE_PER_FILE, ATBU_PERSIST_TYPE_PER_DIR],
        "pb:": [ATBU_PERSIST_TYPE_PER_FILE, ATBU_PERSIST_TYPE_PER_DIR],
    }
    current_persist_types = [ATBU_PERSIST_TYPE_PER_DIR]
    locations = []
    for loc in arguments:
        orig_loc = loc
        if max_allowed is not None and len(locations) >= max_allowed:
            raise InvalidCommandLineArgument(
                f"You have already specified {len(locations)} locations. "
                f"The argument '{loc}' is invalid."
            )
        if loc in per_arg_to_persist_types:
            current_persist_types = per_arg_to_persist_types[loc]
            continue
        if len(loc) == 0:
            raise InvalidCommandLineArgument(
                f"Invalid argument '{orig_loc}'. Expected a directory but got '{loc}'"
            )
        loc = os.path.abspath(loc)
        if not os.path.exists(loc):
            raise InvalidCommandLineArgument(
                f"The specified location does not exist: {loc}"
            )
        locations.append(
            (
                loc,
                current_persist_types,
            )
        )
    if len(locations) < min_required:
        raise InvalidCommandLineArgument(
            f"You have only specified '{len(locations)}' which is less than the {min_required} required."
        )
    if max_allowed is not None and len(locations) > max_allowed:
        # Given above checks, this should never occur.
        raise InvalidCommandLineArgument(
            f"You specified {len(locations)} locations which is more than the {max_allowed} allowed."
        )
    return locations


@dataclass
class LocationSummaryItem:
    location: str
    total_all_files: int
    total_found_unique_files: int
    total_found_physical_files: int
    total_skipped_files: int


def log_location_summary_item(lsi: LocationSummaryItem):
    logging.info(f"Location .............................. {lsi.location}")
    logging.info(f"Location total all files .............. {lsi.total_all_files}")
    logging.info(
        f"Location total found unique files ..... {lsi.total_found_unique_files}"
    )
    logging.info(
        f"Location total found physical files ... {lsi.total_found_physical_files}"
    )
    if lsi.total_skipped_files > 0:
        logging.warning(
            f"Location total skipped files ....... {lsi.total_skipped_files}"
        )
    else:
        logging.info(
            f"Location total skipped files .......... {lsi.total_skipped_files}"
        )


def handle_savedb(args):
    logging.debug("handle_savedb func")
    if args.locations is None:
        raise InvalidCommandLineArgument("You must specify one or more locations.")
    logging.debug(f"locations={args.locations}")
    location_persist_types = extract_location_info(
        args.locations, min_required=1, max_allowed=None
    )

    if not hasattr(args, "database"):
        raise InvalidCommandLineArgument(
            "The --database location has not been specified."
        )
    database_path = args.database
    database_skipped_path = os.path.splitext(database_path)[0] + "-skipped.txt"
    logging.info(f"Database: {database_path}")
    if os.path.exists(database_path):
        if not args.overwrite:
            raise InvalidCommandLineArgument(
                f"The database file already exists: {database_path}"
            )
        os.remove(database_path)
        logging.info(f"Database file erased: {database_path}")
    if os.path.exists(database_skipped_path):
        if not args.overwrite:
            raise InvalidCommandLineArgument(
                f"The database skipped list file already exists: {database_skipped_path}"
            )
        os.remove(database_skipped_path)
        logging.info(f"Database skipped file erased: {database_skipped_path}")
    all_locations_info: defaultdict[str, list[FileInformationPersistent]] = defaultdict(
        list[FileInformationPersistent]
    )
    all_locations_info_skipped = []
    location_summary = []
    for lp in location_persist_types:
        location = lp[0]
        persist_types = lp[1]
        if not os.path.exists(location):
            raise FileNotFoundError(f"Directory does not exist: {location}")

        location_DBs = FileInformationDatabaseCollection(
            source_path=location, persist_types=persist_types
        )
        updater = location_DBs.update(
            change_detection_type=args.change_detection_type,
            update_stale=args.update_stale,
            whatif=False,
        )
        location_info_found = location_DBs.get_dict_digest_to_fi()
        location_info_skipped = updater.skipped_files

        lsi = LocationSummaryItem(
            location=location,
            total_all_files=len(location_info_found) + len(location_info_skipped),
            total_found_unique_files=len(location_info_found),
            total_found_physical_files=sum(
                len(l) for l in location_info_found.values()
            ),
            total_skipped_files=len(location_info_skipped),
        )
        location_summary.append(lsi)
        log_location_summary_item(lsi=lsi)
        for k, v in location_info_found.items():
            all_locations_info[k] += v
        all_locations_info_skipped += location_info_skipped
    save_database_from_dict_file_info(
        dict_digest_to_file_info_list=all_locations_info, database_path=database_path
    )
    with open(database_skipped_path, "w", encoding="utf-8") as database_skipped_file:
        database_skipped_file.write(json.dumps(all_locations_info_skipped))
    logging.info(
        f"========================================================================="
    )
    logging.info(f"The following is a recap of summary information output above:")
    for lsi in location_summary:
        log_location_summary_item(lsi=lsi)
        logging.info(
            f"-------------------------------------------------------------------------"
        )
    logging.info(f"All locations total unique files ...... {len(all_locations_info)}")
    logging.info(
        f"All locations total physical files .... {sum(len(l) for l in all_locations_info.values())}"
    )
    if len(all_locations_info_skipped) > 0:
        logging.warning(
            f"All locations skipped files ........... {len(all_locations_info_skipped)}"
        )
    else:
        logging.info(
            f"All locations skipped files ........... {len(all_locations_info_skipped)}"
        )



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Persistent file information diff'ing.
"""

import os
import logging
import re
import time

from ..common.constants import *
from ..common.exception import *
from ..common.util_helpers import is_platform_path_case_sensitive
from .database import FileInformationDatabaseCollection, extract_location_info
from ..common.file_info import CHANGE_DETECTION_TYPE_DATESIZE, FileInformationPersistent
from ..common.hasher import GlobalHasherDefinitions
from ..common.mp_global import get_verbosity_level

DIFF_COMMAND_MOVE_DUPLICATES = "move-duplicates"
DIFF_COMMAND_REMOVE_DUPLICATES = "remove-duplicates"
DIFF_COMMAND_CHOICES = [
    DIFF_COMMAND_MOVE_DUPLICATES,
    DIFF_COMMAND_REMOVE_DUPLICATES,
]


def remove_empty_directorires(
    root_dir_paths: set[str],
    whatif: bool,
):
    """Remove empty directories within the specified root_dir_paths."""
    whatif_str = "(what_if)" if whatif else ""
    directory_removal_attempts = 3
    retry_directory_removal = True
    while directory_removal_attempts > 0 and retry_directory_removal:
        logging.info(f"Removing empty directories...")
        retry_directory_removal = False
        directory_removal_attempts -= 1
        for directory_affected in root_dir_paths:
            if os.path.isdir(directory_affected):
                try:
                    if not whatif:
                        if len(os.listdir(directory_affected)) == 0:
                            os.removedirs(directory_affected)
                    logging.info(
                        f"Successfully removed{whatif_str} {directory_affected}"
                    )
                except OSError:
                    retry_directory_removal = True
                    logging.info(f"Failed to remove {directory_affected}")
        if retry_directory_removal and directory_removal_attempts >= 1:
            logging.info(f"Some directories were not removed, trying again.")
            logging.info(f"If issue persists, check directory permissions/attributes.")
            time.sleep(2)
    directories_removed = 0
    for directory_affected in root_dir_paths:
        if not os.path.isdir(directory_affected):
            directories_removed += 1
    return directories_removed


class FileInformationCommandBase:
    def __init__(
        self,
        root_source_dir: str,
        file_info_to_affect: list[FileInformationPersistent],
        whatif: bool = False,
    ):
        self.whatif = whatif
        if root_source_dir is None or not os.path.isdir(root_source_dir):
            raise LocationDoesNotExistException(
                f"The location is not a directory: {root_source_dir}"
            )
        self.root_source_dir = root_source_dir
        self._file_info_to_affect = file_info_to_affect
        self.directories_affected = set()
        if self.whatif:
            self.whatif_str = "(--whatif) "
        else:
            self.whatif_str = ""
        self.unique_files_affected = 0
        self.physical_files_total = 0
        self.physical_files_affected = 0
        self.config_files_total = 0
        self.config_files_affected = 0
        self.directories_removed = 0

    @property
    def file_info_to_affect(self) -> list[FileInformationPersistent]:
        """The file information upon which to perform the comand."""
        return self._file_info_to_affect

    @file_info_to_affect.setter
    def file_info_to_affect(self, value: list[FileInformationPersistent]):
        if value is None:
            raise ValueError(f"file_info_to_affect must be specified.")
        self._file_info_to_affect = value


class MoveFileInformationCommand(FileInformationCommandBase):
    def __init__(
        self,
        root_source_dir: str,
        file_info_to_affect: list[FileInformationPersistent],
        move_destination: str,
        whatif: bool = False,
    ):
        super().__init__(
            root_source_dir=root_source_dir,
            file_info_to_affect=file_info_to_affect,
            whatif=whatif,
        )
        if move_destination is None:
            raise ValueError(
                "move_destination is None but must be an absolute path to an existing directory."
            )
        if os.path.abspath(move_destination).strip("\\/") != move_destination.strip(
            "\\/"
        ):
            raise ValueError(
                "move_destination must be an absolute path to an existing directory."
            )
        self.move_destination = move_destination

    def _get_file_info_move_dest_path(
        self,
        file_info: FileInformationPersistent,
        root_source_dir: str,
        root_dest_dir: str,
    ):
        flags = 0
        if not is_platform_path_case_sensitive():
            flags = re.IGNORECASE
        actual_root_source_dir = os.path.commonpath(
            [file_info.dirname, root_source_dir]
        )
        rel_file_info_path = re.sub(
            f"^{re.escape(actual_root_source_dir)}", "", file_info.dirname, flags=flags
        ).lstrip("\\/")
        return os.path.join(
            root_dest_dir, rel_file_info_path, file_info.basename
        ), os.path.join(root_dest_dir, rel_file_info_path, file_info.config_basename)

    def _move_file_to_dest_path(
        self,
        file_info: FileInformationPersistent,
        root_source_dir: str,
        root_dest_dir: str,
        whatif: bool,
    ):
        """Move a file and its FileInformationPersistent from a source path to a destination path, retaining the
        same directory structure for non-root portions of the paths.
        """
        file_moved = False
        config_file_moved = False
        (
            destination_file_path,
            destination_config_file_path,
        ) = self._get_file_info_move_dest_path(
            file_info, root_source_dir, root_dest_dir
        )
        if file_info.is_loaded_from_db:
            destination_config_file_path = None
        logging.info(
            f"Moving{self.whatif_str} {file_info.path} ---to--> {destination_file_path} digest={file_info.get_digest()}"
        )
        if not whatif:
            os.makedirs(os.path.split(destination_file_path)[0], exist_ok=True)
            os.rename(file_info.path, destination_file_path)
            file_moved = True
            if not file_info.is_loaded_from_db:
                logging.info(
                    f"Moving{self.whatif_str} {file_info.info_data_file_path} ---to--> {destination_config_file_path}"
                )
                try:
                    os.makedirs(
                        os.path.split(destination_config_file_path)[0], exist_ok=True
                    )
                    os.rename(
                        file_info.info_data_file_path, destination_config_file_path
                    )
                    config_file_moved = True
                except Exception as ex:
                    try:
                        # Undo the successful user data file move.
                        os.rename(src=destination_file_path, dst=file_info.path)
                    except Exception:
                        pass
                    logging.error(
                        f"Failure moving config file: path={file_info.info_data_file_path} {exc_to_string(ex)}"
                    )
                    raise
            if get_verbosity_level() > 0:
                logging.info(f"Move successful.")
        else:
            logging.info(
                f"Moving{self.whatif_str} {file_info.path} ---to--> {destination_file_path} digest={file_info.get_digest()}"
            )
            logging.info(
                f"Moving{self.whatif_str} {file_info.info_data_file_path} ---to--> {destination_config_file_path}"
            )
            if get_verbosity_level() > 0:
                logging.info(f"Move successful.")
            file_moved = True
            config_file_moved = True
        return file_moved, config_file_moved

    def perform_command(self):
        """Move relevant files and associated config files from a source tree to a destination tree."""

        logging.info(
            f"======================================================================"
        )
        logging.info(
            f"                                MOVING                                "
        )
        logging.info(
            f"======================================================================"
        )
        logging.info(f"Moving duplicates in Location A: {self.root_source_dir}")

        unique_file_counter_hash_set = set()
        file_info_to_process: FileInformationPersistent
        for file_info_to_process in self.file_info_to_affect:
            unique_file_counter_hash_set.add(file_info_to_process.get_digest())
            self.physical_files_total += 1
            is_file_moved, is_config_file_moved = self._move_file_to_dest_path(
                file_info_to_process,
                self.root_source_dir,
                self.move_destination,
                whatif=self.whatif,
            )
            if is_file_moved:
                self.directories_affected.add(file_info_to_process.dirname)
                self.physical_files_affected += 1
            self.config_files_total += 1
            if is_config_file_moved:
                self.config_files_affected += 1
        self.unique_files_affected += len(unique_file_counter_hash_set)
        self.directories_removed += remove_empty_directorires(
            root_dir_paths=self.directories_affected,
            whatif=self.whatif,
        )


class RemoveFileInformationCommand(FileInformationCommandBase):
    def __init__(
        self,
        root_source_dir: str,
        file_info_to_affect: list[FileInformationPersistent],
        whatif: bool = False,
    ):
        super().__init__(
            root_source_dir=root_source_dir,
            file_info_to_affect=file_info_to_affect,
            whatif=whatif,
        )

    def _log_remove_error(
        self,
        is_config_remove_error: bool,
        file_info: FileInformationPersistent,
        exception: Exception,
    ) -> bool:
        """The a removal error, returns True if error should be raised, else do not raise."""
        file_path = file_info.path
        if is_config_remove_error:
            file_path = file_info.info_data_file_path
        logging.error(f"Removal failed: {file_path} error={exception}")
        if isinstance(exception, OSError):
            logging.error(
                f"Ensure the file is read/write, is not held/locked, and that you have permission to remove the file."
            )
        elif isinstance(exception, FileNotFoundError):
            logging.error(f"The file was not found.")
        elif isinstance(exception, IsADirectoryError):
            logging.error(f"The path is a directory not a file.")
        else:
            logging.error(f"Unexpected exception.")
            return True
        if is_config_remove_error:
            logging.error(f"The config file could not be removed, raising exception.")
            return True  # Raise the exception.
        return False  # Do not raise the exception.

    def _remove_file(
        self,
        is_remove_config: bool,
        file_info_to_remove: FileInformationPersistent,
    ):
        if is_remove_config and file_info_to_remove.is_loaded_from_db:
            return
        path_to_remove = (
            file_info_to_remove.path
            if not is_remove_config
            else file_info_to_remove.info_data_file_path
        )
        if not is_remove_config:
            logging.info(
                f"Removing{self.whatif_str} {path_to_remove} digest={file_info_to_remove.primary_digest}"
            )
        else:
            logging.info(f"Removing{self.whatif_str} {path_to_remove}")

        try:
            os.remove(path_to_remove)
            if get_verbosity_level() > 0:
                logging.info(f"Remove successful.")
        except OSError as err:
            raise_exception = self._log_remove_error(
                is_config_remove_error=is_remove_config,
                file_info=path_to_remove,
                exception=err,
            )
            if raise_exception:
                raise
            # TODO: Add switch for selection to try to remove read only (but not make that default)
            # logging.warning(f"    OSError: {err}")
            # logging.warning(f"    Trying to remove read-only attribute: chmod(""{file_info_to_remove.path}"",stat.S_IWRITE)")
            # os.chmod(file_info_to_remove.path, stat.S_IWRITE)
            # os.remove(file_info_to_remove.path)
            # file_removed = True
        except BaseException as err:
            raise_exception = self._log_remove_error(
                is_config_remove_error=False,
                file_info=file_info_to_remove,
                exception=err,
            )
            if raise_exception:
                raise

    def _remove_file_and_file_info(
        self, file_info_to_remove: FileInformationPersistent, whatif: bool
    ) -> tuple[bool, bool]:
        file_removed = False
        config_file_removed = False
        is_loaded_from_db = file_info_to_remove.is_loaded_from_db
        if not is_loaded_from_db and not os.path.isfile(
            file_info_to_remove.info_data_file_path
        ):
            raise InvalidStateError(
                f"Unexpected state: {ATBU_ACRONUM_U} configuration file note found: {file_info_to_remove.info_data_file_path}"
            )
        if not whatif:
            # Remove main user data file.
            self._remove_file(
                is_remove_config=False,
                file_info_to_remove=file_info_to_remove,
            )
            file_removed = True

            if not is_loaded_from_db:
                # Remove .atbu config file.
                self._remove_file(
                    is_remove_config=True,
                    file_info_to_remove=file_info_to_remove,
                )
                config_file_removed = True
        else:
            # For 'whatif' case, send feedback of successful removal if the file exists.
            if os.path.isfile(file_info_to_remove.path):
                logging.info(
                    f"Removing{self.whatif_str} {file_info_to_remove.path} digest={file_info_to_remove.get_digest()}"
                )
                if not is_loaded_from_db:
                    logging.info(
                        f"Removing{self.whatif_str} {file_info_to_remove.info_data_file_path}"
                    )
                    if get_verbosity_level() > 0:
                        logging.info(f"Remove successful.")
                file_removed = True
                config_file_removed = True
            else:
                logging.error(f"    File not found error.")
        return file_removed, config_file_removed

    def perform_command(self):
        logging.info(
            f"======================================================================"
        )
        logging.info(
            f"                                REMOVING                              "
        )
        logging.info(
            f"======================================================================"
        )
        logging.info(f"Removing duplicates in Location A: {self.root_source_dir}")
        unique_file_counter_hash_set = set()
        file_info_to_process: FileInformationPersistent
        for file_info_to_process in self.file_info_to_affect:
            unique_file_counter_hash_set.add(file_info_to_process.get_digest())
            self.physical_files_total += 1
            is_file_removed, is_config_files_removed = self._remove_file_and_file_info(
                file_info_to_remove=file_info_to_process,
                whatif=self.whatif,
            )
            if is_file_removed:
                self.directories_affected.add(file_info_to_process.dirname)
                self.physical_files_affected += 1
            self.config_files_total += 1
            if is_config_files_removed:
                self.config_files_affected += 1
        self.unique_files_affected += len(unique_file_counter_hash_set)
        self.directories_removed += remove_empty_directorires(
            root_dir_paths=self.directories_affected,
            whatif=self.whatif,
        )


def is_file_info_list_bad_state(
    primary_hasher_name, file_info_list: list[FileInformationPersistent]
):
    primary_digest = file_info_list[0].get_digest(primary_hasher_name)
    size_in_bytes = file_info_list[0].size_in_bytes
    state_error = False
    for file_info in file_info_list[1:]:
        other_primary_digest = file_info.get_digest(primary_hasher_name)
        if primary_digest != other_primary_digest:
            logging.error(
                f"Unexpected list state error, digest mismatch '{primary_digest}' != '{other_primary_digest}'"
            )
            for file_info2 in file_info_list:
                logging.error(f"item related to bad state: {file_info2}")
            state_error = True
            break
        if size_in_bytes != file_info.size_in_bytes:
            logging.error(
                "Unexpected size mismatch encountered, potential collision, skipping the following file:"
            )
            for file_info2 in file_info_list:
                logging.error(f"Collision item: {file_info2}")
            state_error = True
            break
    return state_error


def rel_path(top_level_dir: str, path: str):
    check_top = os.path.normcase(path[: len(top_level_dir)])
    if os.path.normcase(top_level_dir) != check_top:
        raise InvalidFunctionArgument(
            f"The path must be a subdirectory of top_level_dir: "
            f"top_level_dir={top_level_dir} path={path}"
        )
    rpath = path[len(top_level_dir) :]
    if rpath[0] in ["\\", "/"]:
        rpath = rpath[1:]
    return rpath


def diff_locations(
    locationA_path: str,
    locationA_info: dict[str, list[FileInformationPersistent]],
    locationB_path: str,
    locationB_info: dict[str, list[FileInformationPersistent]],
    enforce_rel_path_match: bool,
):
    primary_hasher_name = GlobalHasherDefinitions().get_primary_hashing_algo_name()
    logging.info(f"Logging A unique objects ..... {len(locationA_info)}")
    logging.info(f"Logging B unique objects ..... {len(locationB_info)}")
    if len(locationA_info) == 0:
        logging.warning(f"No file info found for Location A")
    if len(locationB_info) == 0:
        logging.warning(f"No file info found for Location B")
    # Start with all items in Location A being in Location B, prune below,
    # whatever is left is result.
    locA_not_in_locB = dict(locationA_info)
    locA_in_locB = {}
    for digest, file_info_list_locA in locationA_info.items():
        if is_file_info_list_bad_state(
            primary_hasher_name=GlobalHasherDefinitions().get_primary_hashing_algo_name(),
            file_info_list=file_info_list_locA,
        ):
            continue
        primary_digest_locA = file_info_list_locA[0].primary_digest
        if primary_digest_locA is None:
            logging.error(f"No primary digest found for")
            continue
        if digest != primary_digest_locA:
            raise Exception(
                f"Key digest does not match item in list. {digest} {primary_digest_locA}"
            )
        file_info_list_locB = locationB_info.get(primary_digest_locA)
        if file_info_list_locB is None:
            logging.debug(
                f"Location A file_info not in Location B: {primary_hasher_name} digest={primary_digest_locA} file_info[0]={file_info_list_locA[0].path}"
            )
            continue
        if is_file_info_list_bad_state(primary_hasher_name, file_info_list_locB):
            logging.warning(
                f"Location B file_info list bad state causes location A file_info to be considered not in B."
            )
            continue
        # Sanity check
        primary_digest_locB = file_info_list_locB[0].primary_digest
        if primary_digest_locA != primary_digest_locB:
            logging.error(
                f"Location A and B primary digests do not match when expected: {primary_hasher_name}_locA={primary_digest_locA} {primary_hasher_name}_locB={primary_digest_locB} pathA={file_info_list_locA[0].path} pathB={file_info_list_locB[0].path}"
            )
            continue
        logging.debug(
            f"Location A and B digests match: {primary_hasher_name}_locA={primary_digest_locA} {primary_hasher_name}_locB={primary_digest_locB} pathA={file_info_list_locA[0].path} pathB={file_info_list_locB[0].path}"
        )
        logging.info(
            f"Location A and B digests match: {primary_hasher_name}={primary_digest_locA} {file_info_list_locA[0].basename}"
        )
        if enforce_rel_path_match:
            is_all_locA_matching = True
            for fi_a in file_info_list_locA:
                is_match_found = False
                rp_a = rel_path(top_level_dir=locationA_path, path=fi_a.nc_path)
                for fi_b in file_info_list_locB:
                    rp_b = rel_path(top_level_dir=locationB_path, path=fi_b.nc_path)
                    if rp_a == rp_b:
                        is_match_found = True
                        break
                if not is_match_found:
                    is_all_locA_matching = False
                    logging.debug(
                        f"Location A relative path not found in location B: "
                        f"pathA={rp_a}"
                    )
                    for fi_b in file_info_list_locB:
                        rp_b = rel_path(top_level_dir=locationB_path, path=fi_b.nc_path)
                        logging.debug(f"    Location B non-match: pathB={rp_b}")
                    break
            if not is_all_locA_matching:
                # Not considered found in both A and B given enforce_rel_path_match.
                continue
        # Item found in both Location A and B
        # Transfer item from "not in" dict to "in" dict, then delete from "not in" dict.
        locA_in_locB[primary_digest_locA] = locA_not_in_locB[primary_digest_locA]
        del locA_not_in_locB[primary_digest_locA]
    if len(locA_not_in_locB) == 0:
        logging.info("All items in Location A were found in Location B")
    else:
        logging.info(
            f"======================================== RESULTS ============================================="
        )
        logging.info(f"Files in Location A *not* found in Location B:")
        for digest, file_info_list in locA_not_in_locB.items():
            for file_info in file_info_list:
                logging.info(f"File in A *not* in B: {file_info.path}")
                if get_verbosity_level() > 0:
                    logging.info(file_info)
            logging.info("----------------------------------------")
    return locA_not_in_locB, locA_in_locB


def handle_diff(args):
    locations = extract_location_info(args.locations, min_required=2, max_allowed=2)
    locationA = locations[0][0]
    locationA_persist_types = locations[0][1]
    locationB = locations[1][0]
    locationB_persist_types = locations[1][1]

    locationA_DBs = FileInformationDatabaseCollection(
        source_path=locationA, persist_types=locationA_persist_types
    )

    locationB_DBs = FileInformationDatabaseCollection(
        source_path=locationB, persist_types=locationB_persist_types
    )

    logging.info(f"{'Location A ':.<40} {locationA}")
    logging.info(f"{'Location A persist types ':.<40} {locationA_persist_types}")
    logging.info(f"{'Location B ':.<40} {locationB}")
    logging.info(f"{'Location B persist types ':.<40} {locationB_persist_types}")

    whatif_str = " (--whatif)" if args.whatif else " "

    post_diff_command = None
    command_line_action_str = None
    if hasattr(args, "action"):
        command_line_action_str = args.action
        logging.debug(f"action_str={command_line_action_str}")
    if command_line_action_str == "move-duplicates":
        past_tense_verb = "moved"
        if not hasattr(args, "move_destination") or not args.move_destination:
            raise InvalidCommandLineArgument(
                "The --move-destination location has not been specified."
            )
        post_diff_command = MoveFileInformationCommand(
            root_source_dir=locationA,
            file_info_to_affect=None,
            move_destination=args.move_destination,
            whatif=args.whatif,
        )
    elif command_line_action_str == "remove-duplicates":
        past_tense_verb = "removed"
        if not os.path.isdir(locationA):
            raise InvalidCommandLineArgument(
                f"When --remove-duplicates is specified, "
                f"location A must be a directory. "
                f"Duplicates are removed from location A."
            )
        post_diff_command = RemoveFileInformationCommand(
            root_source_dir=locationA,
            file_info_to_affect=None,
            whatif=args.whatif,
        )
    elif command_line_action_str is None:
        pass
    else:
        raise InvalidCommandLineArgument(
            f"Unknown action specified: {command_line_action_str}"
        )

    logging.info(f"-" * 65)
    logging.info(f"Searching location A: {locationA}")
    updaterA = locationA_DBs.update(
        change_detection_type=args.change_detection_type,
        update_stale=args.update_stale,
        whatif=args.whatif,
    )
    locationA_info = locationA_DBs.get_dict_digest_to_fi()
    locationA_skipped = updaterA.skipped_files

    logging.info(f"-" * 65)
    logging.info(f"Searching location B: {locationB}")
    updaterB = locationB_DBs.update(
        change_detection_type=args.change_detection_type,
        update_stale=args.update_stale,
        whatif=args.whatif,
    )
    locationB_info = locationB_DBs.get_dict_digest_to_fi()
    locationB_skipped = updaterB.skipped_files

    locA_not_in_locB: dict[str, list[FileInformationPersistent]]
    locA_in_locB: dict[str, list[FileInformationPersistent]]
    locA_not_in_locB, locA_in_locB = diff_locations(
        locationA_path=locationA_DBs.source_path,
        locationA_info=locationA_info,
        locationB_path=locationB_DBs.source_path,
        locationB_info=locationB_info,
        enforce_rel_path_match=args.relpath_match,
    )

    locationA_skipped_item: tuple[FileInformationPersistent, str]
    for locationA_skipped_item in locationA_skipped:
        logging.warning(
            f"Skipped: {locationA_skipped_item[1]}: {locationA_skipped_item[0].path}"
        )
    locationB_skipped_item: tuple[FileInformationPersistent, str]
    for locationB_skipped_item in locationB_skipped:
        logging.warning(
            f"Skipped: {locationB_skipped_item[1]}: {locationB_skipped_item[0].path}"
        )

    try:
        if post_diff_command is not None:
            locA_in_locB_flattened = sorted(
                [file_info for v in locA_in_locB.values() for file_info in v],
                key=lambda x: x.path,
            )
            post_diff_command.file_info_to_affect = locA_in_locB_flattened
            post_diff_command.perform_command()

            if not args.whatif:
                # After performing the command, update the DBs.
                # Use CHANGE_DETECTION_TYPE_DATESIZE regardless of user
                # choice because this update is merely to filter out
                # what no longer exists from per-dir .json DB files.
                logging.info(f"Starting post-command location A update...")
                post_command_updaterA = locationA_DBs.update(
                    change_detection_type=CHANGE_DETECTION_TYPE_DATESIZE,
                    update_stale=args.update_stale,
                    whatif=args.whatif,
                )
                if len(post_command_updaterA.skipped_files) > 0:
                    logging.warning(
                        f"{'Total Location A post-command skipped files ':.<65}"
                        f" {len(post_command_updaterA.skipped_files)}"
                    )
    except Exception as ex:
        logging.info("")
        logging.error(
            f"ERROR: The post-diff command failed: {os.linesep}{exc_to_string_with_newlines(ex)}"
        )
        logging.info("")
    finally:
        if len(updaterA.sneaky_corruption_potentials) > 0:
            logging.info(f"=" * 65)
            logging.info(f"Potential sneaky corruption Location A: {locationA}")
            for scp in updaterA.sneaky_corruption_potentials:
                logging.info(f"        path={scp.file_info.path}")
                logging.info(f"        old_size={scp.old_size_in_bytes}")
                logging.info(f"        cur_size={scp.cur_size_in_bytes}")
                logging.info(f"        old_time={scp.old_modified_time}")
                logging.info(f"        cur_time={scp.cur_modified_time}")
                logging.info(f"        old_digest={scp.old_digest}")
                logging.info(f"        cur_digest={scp.cur_digest}")
                logging.info(f"-" * 65)
            logging.info(
                f"    Total potential sneaky corruption Location A: {len(updaterA.sneaky_corruption_potentials)}"
            )
            logging.info(f"=" * 65)
        if len(updaterB.sneaky_corruption_potentials) > 0:
            logging.info(f"=" * 65)
            logging.info(f"Potential sneaky corruption Location B: {locationB}")
            for scp in updaterB.sneaky_corruption_potentials:
                logging.info(f"        path={scp.file_info.path}")
                logging.info(f"        old_size={scp.old_size_in_bytes}")
                logging.info(f"        cur_size={scp.cur_size_in_bytes}")
                logging.info(f"        old_time={scp.old_modified_time}")
                logging.info(f"        cur_time={scp.cur_modified_time}")
                logging.info(f"        old_digest={scp.old_digest}")
                logging.info(f"        cur_digest={scp.cur_digest}")
                logging.info(f"-" * 65)
            logging.info(
                f"{'Total potential sneaky corruption Location B:':.<65} {len(updaterB.sneaky_corruption_potentials)} (see details above)"
            )
            logging.info(f"=" * 65)
        logging.info(f"{'Location A ':.<65} {locationA}")
        logging.info(f"{'Location B ':.<65} {locationB}")
        logging.info(f"{'Total Location A unique files ':.<65} {len(locationA_info)}")
        logging.info(
            f"{'Total Location A skipped files ':.<65} {len(locationA_skipped)}"
        )
        logging.info(f"{'Total Location B unique files ':.<65} {len(locationB_info)}")
        logging.info(
            f"{'Total Location B skipped files ':.<65} {len(locationB_skipped)}"
        )
        logging.info(
            f"{'Total Location A unique files also in Location B ':.<65} {len(locA_in_locB)}"
        )
        logging.info(
            f"{'Total Location A unique files not found in Location B ':.<65} {len(locA_not_in_locB)}"
        )
        if post_diff_command is not None:
            logging.info(f"Summary '{command_line_action_str}'...")
            whatif_str = " (--whatif)" if post_diff_command.whatif else " "
            past_tense_verb_whatif = f"{past_tense_verb}{whatif_str}"
            logging.info(
                f"{'Total Location A unique files ' + past_tense_verb_whatif:.<65} {post_diff_command.unique_files_affected}"
            )
            logging.info(
                f"{'Total Location A physical files ' + past_tense_verb_whatif:.<65} {post_diff_command.physical_files_affected}"
            )
            if locationA_DBs.has_per_file_persistence:
                logging.info(
                    f"{'Total Location A config files ' + past_tense_verb_whatif:.<65} {post_diff_command.config_files_affected}"
                )
                logging.info(
                    f"{'Total Location A config files not ' + past_tense_verb_whatif:.<65} {post_diff_command.config_files_total - post_diff_command.config_files_affected}"
                )
            logging.info(
                f"{'Total Location A affected directories ':.<65} {len(post_diff_command.directories_affected)}"
            )
            logging.info(
                f"{'Total Location A affected directories emptied/removed ':.<65} {post_diff_command.directories_removed}"
            )



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""AES CBS paddeded encryptor/decryptor. Encrypt a given plaintext such that
the resulting ciphertext is padded to the block size (16-bytes).
"""

from abc import ABC, abstractmethod
from cryptography.hazmat.primitives.ciphers import (
    Cipher,
    algorithms,
    modes,
    CipherContext,
)
from .exception import AlreadyFinalizedError, EncryptionDecryptionFailure


class AES_CBC_Base(ABC):
    """The base for both the encryptor/decryptor. Provides common 'update' functionality."""

    BLOCK_SIZE: int = 16

    def __init__(self, cc: CipherContext, _max_block_retention: int):
        self._buffered_data = bytearray()
        self._cc = cc
        self._num_block_retention = _max_block_retention
        self._num_byte_retention = AES_CBC_Base.BLOCK_SIZE * self._num_block_retention
        self._finished = False

    @property
    def is_finalized(self):
        return self._finished

    @property
    @abstractmethod
    def IV(self):
        pass

    def update(self, input_bytes: bytes) -> bytes:
        return self._process_data(input_bytes)

    def _process_data(self, data: bytes) -> bytes:
        if self._finished:
            raise AlreadyFinalizedError(f"Instance already finalized.")
        output_text = None
        self._buffered_data.extend(data)
        blocks_available = int(len(self._buffered_data) / AES_CBC_Base.BLOCK_SIZE)
        remainder_available = int(len(self._buffered_data) % AES_CBC_Base.BLOCK_SIZE)
        if remainder_available > 0:
            # Given padding, a remaining amount implies an eventual retained block.
            blocks_to_update = blocks_available
        else:
            # No remainder, subtract retention requirement.
            blocks_to_update = blocks_available - self._num_block_retention
        if blocks_to_update <= 0:
            return bytes()
        bytes_to_update = blocks_to_update * AES_CBC_Base.BLOCK_SIZE
        if bytes_to_update > 0:
            output_text = self._cc.update(self._buffered_data[:bytes_to_update])
            del self._buffered_data[:bytes_to_update]
            if len(self._buffered_data) > max(
                self._num_byte_retention, AES_CBC_Base.BLOCK_SIZE - 1
            ):
                raise EncryptionDecryptionFailure(
                    f"Buffered data of {len(self._buffered_data)} bytes is unexpectedly too much."
                )
        return output_text


class AesCbcPaddingEncryptor(AES_CBC_Base):
    """AES CBC padding encryptor."""

    def __init__(self, key: bytes, IV: bytes):
        super().__init__(
            Cipher(algorithms.AES(key), modes.CBC(IV)).encryptor(),
            _max_block_retention=0,
        )
        self._key = key
        self._IV = IV

    @property
    def IV(self):
        return self._IV

    def finalize(self) -> bytes:
        if self._finished:
            raise AlreadyFinalizedError(f"Instance already finalized.")
        self._finished = True
        #
        # For the encryptor, the length of self._buffered_data should be
        # zero to BLOCK_SIZE-1.
        #
        if len(self._buffered_data) >= AES_CBC_Base.BLOCK_SIZE:
            raise EncryptionDecryptionFailure(
                f"Encryption failed: "
                f"Block size {len(self._buffered_data)} is unexpected. "
                f"Expected no more than {AES_CBC_Base.BLOCK_SIZE - 1} bytes."
            )
        padding_needed = AES_CBC_Base.BLOCK_SIZE - len(self._buffered_data)
        for _ in range(0, padding_needed):
            self._buffered_data.append(padding_needed)
        ciphertext = self._cc.update(self._buffered_data) + self._cc.finalize()
        return ciphertext


class AesCbcPaddingDecryptor(AES_CBC_Base):
    """AES CBC padding decryptor."""

    def __init__(self, key: bytes, IV: bytes):
        self._key = key
        self._IV = IV
        super().__init__(
            cc=Cipher(algorithms.AES(self._key), modes.CBC(self._IV)).decryptor(),
            _max_block_retention=1,
        )

    @property
    def IV(self):
        return self._IV

    def finalize(self) -> bytes:
        if self._finished:
            raise AlreadyFinalizedError(f"Instance already finalized.")
        self._finished = True
        if len(self._buffered_data) != self._num_byte_retention:
            raise EncryptionDecryptionFailure(
                f"Decryption failed: "
                f"Final data of {len(self._buffered_data)} bytes is unexpected. "
                f"Expected {self._num_byte_retention} or more bytes."
            )
        plaintext = self._cc.update(self._buffered_data) + self._cc.finalize()
        amount_of_padding = int(plaintext[-1])
        if amount_of_padding > AES_CBC_Base.BLOCK_SIZE:
            raise EncryptionDecryptionFailure(
                f"Decryption failed: "
                f"Last block end byte is {plaintext[-1]} which is an unexpected padding byte."
            )
        first_padding_index = AES_CBC_Base.BLOCK_SIZE - amount_of_padding
        for i in range(first_padding_index, first_padding_index + amount_of_padding):
            if plaintext[i] != amount_of_padding:
                raise EncryptionDecryptionFailure(
                    f"Decryption failed: "
                    f"Expected padding byte value {amount_of_padding} but got {plaintext[i]}"
                )
        plaintext = plaintext[0:first_padding_index]
        return plaintext



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from glob import iglob
from pathlib import Path
from typing import Iterator, Union
from uuid import uuid4
from .base import (
    AUTO_FIND_CREATE_CONTAINER_INDICATOR_CHAR,
    StorageContainerInterface,
    StorageInterface,
    StorageObjectInterface,
)
from ...common.exception import ObjectDoesNotExistError, InstanceRefreshRequired
from ..config import *
from ...common.file_info import FileInformation


def get_filesystem_storage_path(root: Union[Path, str], object_name: Union[Path, str]):
    if isinstance(root, str):
        root = Path(root)
    if isinstance(object_name, str):
        object_name = Path(object_name)
    # Path objects have parent . if just a single component.
    if object_name.parent != Path("."):
        raise ValueError("Object name is not just a name.")
    if len(str(object_name)) < 2:
        return root / object_name
    return root / str(object_name)[:2] / object_name


class FileSystemStorageObjectInterface(StorageObjectInterface):
    def __init__(self, object_path: Path):
        super().__init__()
        self.object_path: Path = convert_to_pathlib_path(object_path)
        self.file_info = FileInformation(path=str(object_path))

    @property
    def size(self):
        try:
            size_in_bytes = self.file_info.size_in_bytes
        except InstanceRefreshRequired:
            self.file_info.refresh_stat_info()
            size_in_bytes = self.file_info.size_in_bytes
        return size_in_bytes

    @property
    def hash(self):
        try:
            digest = self.file_info.primary_digest
        except InstanceRefreshRequired:
            self.file_info.refresh_digests()
            digest = self.file_info.primary_digest
        return digest

    @property
    def name(self):
        return self.object_path.name


class FileSystemStorageContainerInterface(StorageContainerInterface):
    def __init__(self, container_root_path: Path):
        super().__init__()
        self.container_root_path = container_root_path

    @property
    def name(self) -> str:
        return self.container_root_path

    def get_object(self, object_name: str):
        desired_object_path = get_filesystem_storage_path(
            root=self.container_root_path, object_name=object_name
        )
        if not desired_object_path.is_file():
            raise ObjectDoesNotExistError(
                f"The file {str(desired_object_path)} does not exist or is not a file."
            )
        try:
            return FileSystemStorageObjectInterface(object_path=desired_object_path)
        except Exception as ex:
            logging.error(
                f"Unexpected error trying to access the file: {str(desired_object_path)} {exc_to_string(ex)}"
            )
            raise

    def delete_object(self, object_name: str):
        desired_object_path = get_filesystem_storage_path(
            root=self.container_root_path, object_name=object_name
        )
        try:
            desired_object_path.unlink()
        except FileNotFoundError as ex:
            raise ObjectDoesNotExistError(
                f"The object {object_name} does not exist or was not found. {exc_to_string(ex)}"
            ).with_traceback(ex.__traceback__) from ex

    def list_objects(self, prefix: str = None) -> list[StorageObjectInterface]:
        result: list[StorageObjectInterface] = []
        if prefix:
            prefix = os.path.normcase(prefix)
        for p in iglob(os.path.join(self.container_root_path, "**"), recursive=True):
            if prefix:
                just_name = os.path.normcase(os.path.split(p)[1])
                if not just_name.startswith(prefix):
                    continue
            result.append(FileSystemStorageObjectInterface(object_path=p))
        return result


class FileSystemStorageInterface(StorageInterface):
    def __init__(self, storage_def):
        super().__init__()
        self.storage_def: dict = storage_def
        self.provider_id = storage_def[CONFIG_VALUE_NAME_PROVIDER]
        self.backup_root = Path(storage_def[CONFIG_VALUE_NAME_CONTAINER])

    def get_container(self, container_name: str) -> StorageContainerInterface:
        container_name_path = Path(container_name)
        if self.backup_root != container_name_path:
            raise ValueError(
                f"This file system interface only supports accessing one container: container={self.backup_root}"
            )
        return FileSystemStorageContainerInterface(self.backup_root)

    def create_container(self, container_name: str) -> StorageContainerInterface:
        if container_name[-1] == AUTO_FIND_CREATE_CONTAINER_INDICATOR_CHAR:
            base_name = container_name[:-1]
            max_attempts = 1000
            while max_attempts > 0:
                max_attempts -= 1
                candidate = Path(f"{base_name}-{str(uuid4())}")
                if not candidate.exists():
                    candidate.mkdir(parents=True, exist_ok=True)
                    return self.get_container(container_name=candidate)
        else:
            candidate = Path(container_name)
            if candidate.exists():
                raise ContainerAlreadyExistsError(
                    f"FileSystemStorageInterface: The container already exists: {str(candidate)}"
                )
            candidate.mkdir(parents=True, exist_ok=True)
            return self.get_container(container_name=str(candidate))

    def delete_container(self, container_name: str):
        # TBD
        raise NotImplementedError()

    def upload_stream_to_object(
        self,
        container: StorageContainerInterface,
        object_name: str,
        stream,
        source_path: str,
    ):
        if not isinstance(container, FileSystemStorageContainerInterface):
            raise ValueError(
                f"The container is not an instance of FileSystemStorageContainerInterface"
            )
        fsc: FileSystemStorageContainerInterface = container
        if self.backup_root != fsc.container_root_path:
            raise ValueError(
                f"The container does not belong to this interface: caller={fsc.container_root_path} this={fsc.container_root_path}"
            )
        storage_path = get_filesystem_storage_path(
            root=self.backup_root, object_name=object_name
        )
        storage_path.parent.mkdir(parents=True, exist_ok=True)
        with open(str(storage_path), "wb") as storage_file:
            for chunk in stream:
                storage_file.write(chunk)

    def download_object_as_stream(
        self, storage_object: FileSystemStorageObjectInterface, chunk_size: int
    ) -> Iterator[bytes]:
        storage_path = get_filesystem_storage_path(
            root=self.backup_root, object_name=storage_object.name
        )
        with open(str(storage_path), "rb") as storage_file:
            while True:
                b = storage_file.read(chunk_size)
                yield b
                if len(b) == 0:
                    break

    @property
    def retry_exceptions(self) -> tuple:
        return ()



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Recover-related command line handlers.
"""

from concurrent.futures import Future
from concurrent import futures
from concurrent.futures import ALL_COMPLETED
from datetime import datetime
import glob
import os
import logging
import re
from shutil import copy2
import tempfile

from ..common.exception import (
    ANOMALY_KIND_EXCEPTION,
    BackupInformationFileTimestampNotFound,
    BackupInformationRecoveryFailed,
    InvalidCommandLineArgument,
    StorageDefinitionNotFoundError,
    exc_to_string,
)
from ..common.util_helpers import prompt_YN
from .config import (
    AtbuConfig,
)
from .restore import RestoreFile
from .backup_core import (
    BACKUP_OPERATION_NAME_RESTORE,
    ANOMALY_KIND_UNEXPECTED_STATE,
    BACKUP_INFO_EXTENSION,
    BACKUP_INFO_TIME_STAMP_FORMAT,
    BackupAnomaly,
    BackupFileInformation,
    StorageDefinition,
    file_operation_future_result,
    BackupPipelineWorkItem,
    is_qualified_for_operation,
    run_operation_stage,
)
from ..common.mp_global import (
    get_process_pool_exec_init_func,
    get_process_pool_exec_init_args,
)
from ..common.mp_pipeline import (
    MultiprocessingPipeline,
    SubprocessPipelineStage,
)


def sort_backup_info_filename_list(filename_list: list[str]):
    re_match_time_stamp = re.compile(rf"(.*)-(\d{{8}}-\d{{6}}){BACKUP_INFO_EXTENSION}")
    temp_list: list[str] = []
    for filename in filename_list:
        m = re_match_time_stamp.match(string=filename)
        if not m:
            continue
        date_time_stamp_str = m.groups()[1]
        dt = datetime.strptime(date_time_stamp_str, BACKUP_INFO_TIME_STAMP_FORMAT)
        temp_list.append(
            (
                filename,
                dt,
            )
        )
    temp_list.sort(key=lambda t: t[1])
    return list(map(lambda t: t[0], temp_list))


def remove_timestamp_from_backupinfo_filename(filename: str):
    re_match_time_stamp = re.compile(
        rf"(.*)(-\d{{8}}-\d{{6}})({BACKUP_INFO_EXTENSION})"
    )
    m = re_match_time_stamp.match(filename)
    if not m:
        raise BackupInformationFileTimestampNotFound(
            f"The backup information timestamp was not found: {filename}"
        )
    return f"{m.groups()[0]}{m.groups()[2]}"


def handle_restore_backup_info(
    storage_def_name: str,
    atbu_cfg: AtbuConfig = None,
    prompt_if_exists: bool = True,
):
    if atbu_cfg is None:
        atbu_cfg = AtbuConfig.access_default_config()
    backup_info_dir = atbu_cfg.get_backup_info_dir()
    storage_def_dict = atbu_cfg.get_storage_def_with_resolved_secrets_deep_copy(
        storage_def_name=storage_def_name
    )
    sd = StorageDefinition.storage_def_from_dict(
        storage_def_name=storage_def_name, storage_def_dict=storage_def_dict
    )
    existing_backup_info = atbu_cfg.get_backup_info_file_paths(
        storage_def_name=storage_def_name
    )
    if prompt_if_exists and len(existing_backup_info) > 0:
        print(f"Found existing local backup info for {storage_def_name}:")
        for ebi in existing_backup_info:
            print(f"  {ebi}")
        print(
            f"""
The recover operation may need to overwrite some or all of the backup information files
listed above. If you are uncertain, you may want to backup those files before proceeding.
"""
        )
        a = prompt_YN(
            prompt_msg=f"You are about to overwrite local backup information.",
            prompt_question=f"Are you certain you want to overwrite local backup information? ",
            default_enter_ans="n",
        )
        if a != "y":
            print("The local backup information will not be overwritten.")
            print("The recovery will be aborted.")
            return

    sp: MultiprocessingPipeline = None
    try:
        with tempfile.TemporaryDirectory(prefix="atbu_rectmp_") as temp_dir:
            logging.info(f"Restoring backup information...")
            interface = sd.create_storage_interface()
            c = interface.get_container(container_name=sd.container_name)
            restore_file_list: list[RestoreFile] = []
            backup_info_objects = c.list_objects(prefix=sd.storage_def_name)
            for bio in backup_info_objects:
                print("Building file information for storage objects...")
                print(f"    {bio.name}")
                fi_bi = BackupFileInformation(path=str(backup_info_dir / bio.name))
                fi_bi.implicit_refresh_allowed = False
                fi_bi.populate_from_header = True
                fi_bi.size_in_bytes = bio.size  # Use storage size until header read.
                fi_bi.restore_path_override = str(backup_info_dir / bio.name)
                fi_bi.storage_object_name = str(bio.name)
                restore_file_bi = RestoreFile(
                    file_info=fi_bi,
                    dest_root_location=str(backup_info_dir),
                    allow_overwrite=True,
                    storage_def=sd,
                    temp_dir=temp_dir,
                )
                restore_file_list.append(restore_file_bi)
            anomalies: list[BackupAnomaly] = []
            sp = MultiprocessingPipeline(
                name="Recover",
                max_simultaneous_work_items=min(os.cpu_count() // 2, 15),
                stages=[
                    SubprocessPipelineStage(
                        fn_determiner=is_qualified_for_operation,
                        fn_worker=run_operation_stage,
                    )
                ],
                process_initfunc=get_process_pool_exec_init_func(),
                process_initargs=get_process_pool_exec_init_args(),
            )

            fut_list: list[Future] = []
            for restore_file in restore_file_list:
                future = sp.submit(
                    work_item=BackupPipelineWorkItem(
                        operation_name=BACKUP_OPERATION_NAME_RESTORE,
                        file_info=restore_file.file_info,
                        is_qualified=True,
                        operation_runner=restore_file,
                    )
                )
                fut_list.append(future)
            done, not_done = futures.wait(set(fut_list), return_when=ALL_COMPLETED)
            if len(done) != len(fut_list):
                msg = (
                    f"Expected len(fut_list) restores to be completed but got {len(done)} instead. "
                    f"Processing whatever is done but you should resolves this issue for proper recovery. "
                    f"done={len(done)} not_done={len(not_done)}"
                )
                logging.error(msg)
                anomalies.append(
                    BackupAnomaly(
                        kind=ANOMALY_KIND_UNEXPECTED_STATE,
                        message=msg,
                    )
                )
            for f in done:
                fi = file_operation_future_result(
                    f=f, anomalies=anomalies, the_operation="Recover backup info"
                )
                if fi is not None and fi.is_successful:
                    logging.info(f"Successfully restored {fi.path}")
    except Exception as ex:
        msg = f"Unhandled exception: {exc_to_string(ex)}"
        logging.error(msg)
        anomalies.append(
            BackupAnomaly(
                kind=ANOMALY_KIND_EXCEPTION,
                exception=ex,
                message=msg,
            )
        )
    finally:
        if sp is not None:
            sp.shutdown()
            anomalies.extend(sp.anomalies)
    existing_backup_info_pat = (
        backup_info_dir / f"{storage_def_name}*{BACKUP_INFO_EXTENSION}"
    )
    existing_backup_info = glob.glob(pathname=str(existing_backup_info_pat))
    existing_backup_info = sort_backup_info_filename_list(
        filename_list=existing_backup_info
    )
    newest_backup_info = existing_backup_info[-1]
    newest_backup_info_wo_stamp = remove_timestamp_from_backupinfo_filename(
        filename=newest_backup_info
    )
    logging.info(f"Copying the most recent backup information...")
    logging.info(f"  {newest_backup_info}")
    logging.info(f"...to...")
    logging.info(f"  {newest_backup_info_wo_stamp}")
    copy2(src=newest_backup_info, dst=newest_backup_info_wo_stamp)
    logging.info(f"Most recent backup information restored.")


def handle_recover(args):
    logging.debug(f"handle_recover")
    storage_def_name = None
    storage_def_config_filename = None
    for arg in args.storage_def_cred_cfg:
        if os.path.isfile(arg):
            if storage_def_config_filename is not None:
                raise InvalidCommandLineArgument(
                    f"Expecting one filename maximum but got two: "
                    f"1={storage_def_config_filename} 2={arg}"
                )
            storage_def_config_filename = arg
        else:
            if storage_def_name is not None:
                raise InvalidCommandLineArgument(
                    f"Expecting one storage definition name but got two: "
                    f"1={storage_def_name} 2={arg}"
                )
            storage_def_name = arg

    # A storage def name which is a folder is always a filesystem storage...
    if storage_def_name and os.path.isdir(storage_def_name):
        filesystem_dir = os.path.abspath(storage_def_name)
        atbu_cfg, storage_def_name, _ = AtbuConfig.access_filesystem_config(
            storage_location_path=filesystem_dir,
            create_if_not_exist=True,
            prompt_to_create=args.prompt,
        )
        # If a config/cred backup file was supplied, import.
        if storage_def_config_filename is not None:
            storage_def_name = atbu_cfg.restore_storage_def(
                storage_def_new_name=None,
                backup_file_path=storage_def_config_filename,
                prompt_if_exists=args.prompt,
            )
            if storage_def_name is None:
                raise BackupInformationRecoveryFailed(
                    f"Could not recover the configuration from {storage_def_config_filename}"
                )
        if not atbu_cfg.is_storage_def_exists(storage_def_name=storage_def_name):
            raise StorageDefinitionNotFoundError(
                f"The storage definition was not found: {storage_def_name}"
            )
        # Restore backup information.
        handle_restore_backup_info(
            storage_def_name=storage_def_name,
            atbu_cfg=atbu_cfg,
            prompt_if_exists=args.prompt,
        )
    else:
        # Non-filesystem (cloud) case.
        if storage_def_name is not None:
            storage_def_name = storage_def_name.lower()
        atbu_cfg = AtbuConfig.access_default_config()
        if storage_def_config_filename is not None:
            # Import config/cred file.
            storage_def_name = atbu_cfg.restore_storage_def(
                storage_def_new_name=storage_def_name,
                backup_file_path=storage_def_config_filename,
                prompt_if_exists=args.prompt,
            )
            if storage_def_name is None:
                raise BackupInformationRecoveryFailed(
                    f"Could not recover the configuration from {storage_def_config_filename}"
                )
        if not atbu_cfg.is_storage_def_exists(storage_def_name=storage_def_name):
            raise StorageDefinitionNotFoundError(
                f"The storage definition was not found: {storage_def_name}"
            )
        handle_restore_backup_info(
            storage_def_name=storage_def_name,
            atbu_cfg=atbu_cfg,
            prompt_if_exists=args.prompt,
        )
    logging.info(f"Completed.")



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Base for ATBU storage interface abstraction, both decoupling ATBU from
other APIs, and providing ATBU itself a common storage interface for its
focused needs.
"""
# pylint: disable=missing-class-docstring

from abc import ABC, abstractmethod
from socket import gaierror
from http.client import NotConnected, ImproperConnectionState
from typing import Iterator
from ...common.constants import *
from ..config import *
from ...common.exception import BackupException, RetryLimitReached

CHUNK_SIZE_5MB = 5 * 1024 * 1024
CHUNK_SIZE_50MB = 50 * 1024 * 1024
DEFAULT_CHUNK_UPLOAD_SIZE = CHUNK_SIZE_5MB
DEFAULT_CHUNK_DOWNLOAD_SIZE = CHUNK_SIZE_50MB
DEFAULT_MAX_SIMULTANEOUS_FILE_BACKUPS = 5

DEFAULT_RETRY_EXCEPTIONS = (
    OSError,
    gaierror,
    NotConnected,
    ImproperConnectionState,
    RetryLimitReached,
)

AUTO_FIND_CREATE_CONTAINER_INDICATOR_CHAR = "*"


class StorageObjectInterface(ABC):
    def __init__(self):
        pass

    @property
    @abstractmethod
    def size(self):
        return None

    @property
    @abstractmethod
    def hash(self):
        return None

    @property
    @abstractmethod
    def name(self):
        return None


class StorageContainerInterface(ABC):
    def __init__(self):
        pass

    @property
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    def get_object(self, object_name: str) -> StorageObjectInterface:
        pass

    @abstractmethod
    def delete_object(self, object_name: str):
        pass

    @abstractmethod
    def list_objects(self, prefix: str = None) -> list[StorageObjectInterface]:
        pass


class StorageInterface(ABC):
    def __init__(self):
        self._retry_exceptions = DEFAULT_RETRY_EXCEPTIONS

    @property
    def upload_chunk_size(self):
        return DEFAULT_CHUNK_UPLOAD_SIZE

    @property
    def download_chunk_size(self):
        return DEFAULT_CHUNK_DOWNLOAD_SIZE

    @abstractmethod
    def get_container(self, container_name: str) -> StorageContainerInterface:
        pass

    @abstractmethod
    def create_container(self, container_name: str) -> StorageContainerInterface:
        pass

    @abstractmethod
    def delete_container(self, container_name: str):
        pass

    @abstractmethod
    def upload_stream_to_object(
        self,
        container: StorageContainerInterface,
        object_name: str,
        stream,
        source_path: str,
    ):
        pass

    @abstractmethod
    def download_object_as_stream(
        self, storage_object: StorageObjectInterface, chunk_size: int
    ) -> Iterator[bytes]:
        pass

    @property
    def retry_exceptions(self):
        return self._retry_exceptions


class StorageInterfaceFactory:
    def __init__(self, storage_def_dict):
        self.storage_def_dict = storage_def_dict

    @property
    def interface_type(self):
        return self.storage_def_dict[CONFIG_VALUE_NAME_INTERFACE_TYPE]

    @property
    def provider_name(self):
        return self.storage_def_dict[CONFIG_VALUE_NAME_PROVIDER]

    def create_storage_interface(self) -> StorageInterface:
        # pylint: disable=import-outside-toplevel
        desired_interface = self.storage_def_dict[CONFIG_VALUE_NAME_INTERFACE_TYPE]
        if desired_interface == CONFIG_INTERFACE_TYPE_LIBCLOUD:
            from .libcloud import LibCloudStorageInterface

            return LibCloudStorageInterface(storage_def=self.storage_def_dict)
        elif desired_interface == CONFIG_INTERFACE_TYPE_GOOGLE:
            from .google import GoogleStorageInterface

            return GoogleStorageInterface(storage_def=self.storage_def_dict)
        elif desired_interface == CONFIG_INTERFACE_TYPE_FILESYSTEM:
            from .filesystem import FileSystemStorageInterface

            return FileSystemStorageInterface(storage_def=self.storage_def_dict)
        else:
            raise BackupException(f"Unknown interface type: {desired_interface}")

    @staticmethod
    def create_factory_from_storage_def_name(storage_def_name: str):
        atbu_cfg = AtbuConfig.access_default_config()
        storage_def_dict = atbu_cfg.get_storage_def_with_resolved_secrets_deep_copy(
            storage_def_name=storage_def_name
        )
        return StorageInterfaceFactory(storage_def_dict=storage_def_dict)



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""ATBU interface for libcloud.
"""
from typing import Iterator
from uuid import uuid4
import libcloud.storage.types
from libcloud.storage.providers import get_driver
import libcloud.storage.base
from libcloud.storage.drivers.azure_blobs import (
    AZURE_DOWNLOAD_CHUNK_SIZE,
    AZURE_UPLOAD_CHUNK_SIZE,
)
from libcloud.utils.retry import RETRY_EXCEPTIONS

from ..credentials import CredentialByteArray
from .base import (
    AUTO_FIND_CREATE_CONTAINER_INDICATOR_CHAR,
    DEFAULT_CHUNK_DOWNLOAD_SIZE,
    DEFAULT_CHUNK_UPLOAD_SIZE,
    StorageContainerInterface,
    StorageInterface,
    StorageObjectInterface,
)
from ...common.exception import ObjectDoesNotExistError, ContainerAlreadyExistsError
from ..config import *


class LibCloudStorageObjectInterface(StorageObjectInterface):
    def __init__(self, storage_object):
        super().__init__()
        self.storage_object: libcloud.storage.base.Object = storage_object

    @property
    def size(self):
        return self.storage_object.size

    @property
    def hash(self):
        return self.storage_object.hash

    @property
    def name(self):
        return self.storage_object.name


class LibCloudStorageContainerInterface(StorageContainerInterface):
    def __init__(self, container: libcloud.storage.base.Container):
        super().__init__()
        self.container = container

    @property
    def name(self) -> str:
        return self.container.name

    def get_object(self, object_name: str):
        try:
            object_interface = LibCloudStorageObjectInterface(
                storage_object=self.container.get_object(object_name)
            )
        except libcloud.storage.types.ObjectDoesNotExistError as ex:
            raise ObjectDoesNotExistError(
                f"The object {object_name} does not exist or was not found. {exc_to_string(ex)}",
                ex,
            ).with_traceback(ex.__traceback__) from ex
        return object_interface

    def delete_object(self, object_name: str):
        try:
            o = self.container.get_object(object_name=object_name)
            self.container.delete_object(o)
        except libcloud.storage.types.ObjectDoesNotExistError as ex:
            raise ObjectDoesNotExistError(
                f"The object {object_name} does not exist or was not found. {exc_to_string(ex)}",
                ex,
            ).with_traceback(ex.__traceback__) from ex

    def list_objects(self, prefix: str = None) -> list[StorageObjectInterface]:
        olist = self.container.list_objects(prefix=prefix)
        result: list[StorageObjectInterface] = []
        for o in olist:
            result.append(LibCloudStorageObjectInterface(storage_object=o))
        return result


class LibCloudStorageInterface(StorageInterface):
    def __init__(self, storage_def):
        super().__init__()
        self.provider_id = storage_def[CONFIG_VALUE_NAME_PROVIDER]
        self.libcloud_driver_class: libcloud.storage.base.StorageDriver = get_driver(
            self.provider_id
        )
        if storage_def[CONFIG_SECTION_DRIVER].get(CONFIG_PASSWORD_TYPE):
            # Remove constructor argument that will be rejected by Libcloud.
            del storage_def[CONFIG_SECTION_DRIVER][CONFIG_PASSWORD_TYPE]
        driver_parameters = dict(storage_def[CONFIG_SECTION_DRIVER])
        secret = driver_parameters[CONFIG_VALUE_NAME_DRIVER_STORAGE_SECRET]
        if isinstance(secret, (CredentialByteArray, bytearray, bytes)):
            secret = secret.decode("utf-8")
        driver_parameters[CONFIG_VALUE_NAME_DRIVER_STORAGE_SECRET] = secret
        self.storage_driver: libcloud.storage.base.StorageDriver = (
            self.libcloud_driver_class(**driver_parameters)
        )

    @property
    def upload_chunk_size(self):
        chunk_size = DEFAULT_CHUNK_UPLOAD_SIZE
        # Add exceptions to the default upload chunk size here.
        if self.provider_id == libcloud.storage.types.Provider.AZURE_BLOBS:
            chunk_size = AZURE_UPLOAD_CHUNK_SIZE
        return chunk_size

    @property
    def download_chunk_size(self):
        chunk_size = DEFAULT_CHUNK_DOWNLOAD_SIZE
        # Add exceptions to the default download chunk size here.
        # Libcloud download_object_as_stream will silently change
        # caller-specified chunk_size without caller knowing, breaking
        # assumption that chunk size will be max until possibly final
        # chunk.
        if self.provider_id == libcloud.storage.types.Provider.AZURE_BLOBS:
            chunk_size = AZURE_DOWNLOAD_CHUNK_SIZE
        return chunk_size

    def get_container(self, container_name: str) -> StorageContainerInterface:
        return LibCloudStorageContainerInterface(
            self.storage_driver.get_container(container_name=container_name)
        )

    def _create_container(self, container_name: str) -> StorageContainerInterface:
        try:
            return LibCloudStorageContainerInterface(
                self.storage_driver.create_container(container_name=container_name)
            )
        except libcloud.storage.types.ContainerAlreadyExistsError as ex:
            raise ContainerAlreadyExistsError(
                f"The storage container {container_name} already exists. {exc_to_string(ex)}"
            ) from ex
        except libcloud.storage.types.InvalidContainerNameError as ex:
            raise InvalidContainerNameError(
                f"The storage container {container_name} is invalid. {exc_to_string(ex)}"
            ) from ex

    def create_container(self, container_name: str) -> StorageContainerInterface:
        if container_name[-1] != AUTO_FIND_CREATE_CONTAINER_INDICATOR_CHAR:
            return self._create_container(container_name=container_name)
        else:
            base_name = container_name[:-1]
            max_attempts = 100
            attempts_remaining = max_attempts
            while True:
                attempts_remaining -= 1
                candidate = f"{base_name}-{str(uuid4())}"
                try:
                    ci = self._create_container(container_name=candidate)
                    return ci
                except (ContainerAlreadyExistsError, InvalidContainerNameError) as ex:
                    if attempts_remaining <= 0:
                        raise ContainerAutoCreationFailed(
                            f"Automatic container creation failed after {max_attempts} attempts. "
                            f"{exc_to_string(ex)}"
                        ) from ex

    def delete_container(self, container_name: str):
        c = self.storage_driver.get_container(container_name=container_name)
        self.storage_driver.delete_container(container=c)

    def upload_stream_to_object(
        self,
        container: StorageContainerInterface,
        object_name: str,
        stream,
        source_path: str,
    ):
        self.storage_driver.upload_object_via_stream(
            iterator=stream.get_iterator_only_proxy(),
            container=container.container,
            object_name=object_name,
        )

    def download_object_as_stream(
        self, storage_object: LibCloudStorageObjectInterface, chunk_size: int
    ) -> Iterator[bytes]:
        return self.storage_driver.download_object_as_stream(
            obj=storage_object.storage_object, chunk_size=chunk_size
        )

    @property
    def retry_exceptions(self) -> tuple:
        return RETRY_EXCEPTIONS



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Credential-related classes.
- CredentialByteArray: Uses byte_array so cred can be zero'ed out afterwards as
  a form of cleanup. Any such efforts at cleanup are *not* security for several
  good reasons, one of which is that CredentialByteArray is not the only party
  handling secrets, where some others such as the cryptography/keyring packages
  use bytes immutable so do not allow, from Python perspective, the zero'ing
  out the memory area they populate with keys/passwords.
- A best effort toward minimizing the time frame when secrets are exposed would be
  to have all parties handling secrets to use a "SecureByteArray" of sorts, which
  would allow memory locking, enforce clearing at cleanup, etc. Stil, None of
  that would remove the need to keep the machine/process secure, but would merely
  minimize the issue caused by immutables which themselves may well leave secrets
  lying around for much longer time periods, if not spreading copies around beyond
  some original.
- Therefore, to achieve security when using a Python process that handles secrets
  in the clear at *any* time for *however* long a duration, the machine and process
  space must be kept clear of threats.
"""
import base64
import os
from typing import Union
from hashlib import pbkdf2_hmac
import keyring
import pwinput

from ..common.util_helpers import is_valid_base64_string

from ..common.aes_cbc import (
    AES_CBC_Base,
    AesCbcPaddingDecryptor,
    AesCbcPaddingEncryptor,
)
from ..common.constants import *
from ..common.exception import (
    CredentialNotFoundError,
    CredentialTypeNotFoundError,
    InvalidBase64StringError,
    PasswordAuthenticationFailure,
    CredentialInvalid,
    CredentialRequestInvalidError,
    CredentialSecretDerivationError,
)


class CredentialByteArray(bytearray):
    """A bytearray with override to allow zeroing out of bytearray
    on demand or when deleted.
    """

    def __del__(self):
        self.zero_array()
        if hasattr(super(), "__del__"):
            super().__del__()

    def zero_array(self):
        for v in [0xCC, 0x55, 0x00]:
            for i, _ in enumerate(self):
                self[i] = v

    def get_portion(self, start_pos=None, end_pos=None):
        if not start_pos:
            start_pos = 0
        if not end_pos:
            end_pos = len(self)
        portion_len = end_pos - start_pos
        r = CredentialByteArray(portion_len)
        idx_r = 0
        for i in range(start_pos, end_pos):
            r[idx_r] = self[i]
            idx_r += 1
        return r

    def to_serialization_dict(self) -> dict:
        return self.hex()

    def from_serialization_dict(self, hex_str):
        self.fromhex(hex_str)


class Credential:

    attr_name_list = [
        "password",
        "salt",
        "pbkdf2",
        "password_auth_hash",
        "key_encryption_key",
        "iv",
        "encrypted_key",
        "the_key",
    ]

    def __init__(
        self,
        password: Union[str, CredentialByteArray] = None,
        salt: CredentialByteArray = None,
        password_auth_hash: CredentialByteArray = None,
        key_encryption_key: CredentialByteArray = None,
        iv: CredentialByteArray = None,
        encrypted_key: CredentialByteArray = None,
        the_key: CredentialByteArray = None,
        key_bit_size=DEFAULT_KEY_BIT_LENGTH,
    ) -> None:

        self.password = None
        self.salt = None
        self.password_auth_hash = None
        self.key_encryption_key = None
        self.iv = None
        self.encrypted_key = None
        self.the_key = None
        self.key_bit_size = None
        self.key_byte_size = None
        self.pbkdf2 = None
        self.work_factor = None

        self.clear(key_bit_size=key_bit_size)
        self.set(
            password=password,
            salt=salt,
            password_auth_hash=password_auth_hash,
            key_encryption_key=key_encryption_key,
            iv=iv,
            encrypted_key=encrypted_key,
            the_key=the_key,
            key_bit_size=key_bit_size,
        )

    @property
    def is_private_key_possible(self):
        """True if a private key is achievable either as is or with a password,
        False if not possible no matter what. See is_password_required.
        Generally, use is_private_key_possible to see if a private is available
        at all by any means. Then, if this is True, use is_password_required to
        determine is the user should be prompted for a password.
        """
        if self.the_key is not None:
            return True
        if self.encrypted_key is None:
            return False
        if self.salt is None or self.iv is None:
            return False
        return True

    @property
    def is_password_required(self):
        """True if password is required, False if not required.
        See is_private_key_possible.
        """
        if self.the_key is not None:
            return False
        if self.key_encryption_key is not None:
            return False
        if self.password is not None:
            return False
        return True

    @property
    def is_private_key_ready(self):
        return self.the_key is not None

    def _set_key_bit_size(self, key_bit_size):
        if key_bit_size is None or key_bit_size not in ALLOWED_KEY_BIT_LENGTHS:
            raise ValueError(
                f"The specified key length of '{key_bit_size} is not one of the "
                f"available choices of {ALLOWED_KEY_BIT_LENGTHS}."
            )
        self.key_bit_size = key_bit_size
        self.key_byte_size = int(key_bit_size / 8)

    def __del__(self):
        self.clear()
        if hasattr(super(), "__del__"):
            super().__del__()

    def clear(self, key_bit_size=DEFAULT_KEY_BIT_LENGTH) -> None:
        for member in Credential.attr_name_list:
            if hasattr(self, member) and isinstance(member, CredentialByteArray):
                cba: CredentialByteArray = getattr(self, member)
                cba.zero_array()
        self.password = None
        self.salt = None
        self.password_auth_hash = None
        self.key_encryption_key = None
        self.iv = None
        self.encrypted_key = None
        self.the_key = None
        self.pbkdf2 = None
        self.work_factor = PBKDF2_WORK_FACTOR
        self._set_key_bit_size(key_bit_size=key_bit_size)

    def clear_password(self):
        if self.password:
            self.password.zero_array()
        self.password = None

    def set(
        self,
        password: Union[str, CredentialByteArray] = None,
        salt: CredentialByteArray = None,
        password_auth_hash: CredentialByteArray = None,
        key_encryption_key: CredentialByteArray = None,
        iv: CredentialByteArray = None,
        encrypted_key: CredentialByteArray = None,
        the_key: CredentialByteArray = None,
        key_bit_size=None,
    ):
        if key_bit_size is not None:
            self._set_key_bit_size(key_bit_size=key_bit_size)

        if isinstance(password, str):
            password = CredentialByteArray(password.encode())
        if password is not None:
            if not isinstance(password, CredentialByteArray):
                raise ValueError("The password is not a CredentialByteArray.")
            self.password = password

        if salt is not None:
            if not isinstance(salt, CredentialByteArray):
                raise ValueError("The salt is not a CredentialByteArray.")
            self.salt = salt

        if password_auth_hash is not None:
            if not isinstance(password_auth_hash, CredentialByteArray):
                raise ValueError("The password_auth_hash is not a CredentialByteArray.")
            self.password_auth_hash = password_auth_hash

        if key_encryption_key is not None:
            if not isinstance(key_encryption_key, CredentialByteArray):
                raise ValueError("The key_encryption_key is not a CredentialByteArray.")
            self.key_encryption_key = password_auth_hash

        if iv is not None:
            if not isinstance(iv, CredentialByteArray):
                raise ValueError("The iv is not a CredentialByteArray.")
            self.iv = iv

        if encrypted_key is not None:
            if not isinstance(encrypted_key, CredentialByteArray):
                raise ValueError("The encrypted_key is not a CredentialByteArray.")
            self.encrypted_key = encrypted_key

        if the_key is not None:
            if not isinstance(the_key, CredentialByteArray):
                raise ValueError("The key is not a CredentialByteArray.")
            self.the_key = the_key

    def create_key(
        self,
        password: Union[str, CredentialByteArray] = None,
        salt: CredentialByteArray = None,
        iv: CredentialByteArray = None,
        the_key: CredentialByteArray = None,
        key_bit_size=DEFAULT_KEY_BIT_LENGTH,
    ) -> None:
        self.clear()
        self._set_key_bit_size(key_bit_size=key_bit_size)

        #
        # Create a password if none was provided.
        #
        if not password:
            password = CredentialByteArray(os.urandom(self.key_byte_size))
        if isinstance(password, str):
            password = CredentialByteArray(password.encode())
        if not isinstance(password, CredentialByteArray):
            raise ValueError(
                f"The password is not either str or bytes. type={type(password)}"
            )
        self.password = password

        #
        # Create a salt if none was provided.
        #
        if not salt:
            salt = CredentialByteArray(os.urandom(self.key_byte_size))
        if not isinstance(salt, CredentialByteArray):
            raise ValueError(f"The salt is not bytes. type={type(salt)}")
        self.salt = salt

        #
        # Create an IV if none was provided.
        #
        if not iv:
            iv = CredentialByteArray(os.urandom(AES_CBC_Base.BLOCK_SIZE))
        if (
            not isinstance(iv, CredentialByteArray)
            or len(iv) != AES_CBC_Base.BLOCK_SIZE
        ):
            raise ValueError(
                f"The salt is not {AES_CBC_Base.BLOCK_SIZE} bytes. type={type(iv)}"
            )
        self.iv = iv

        #
        # Create a key if none was provided.
        #
        if not the_key:
            the_key = CredentialByteArray(os.urandom(self.key_byte_size))
        else:
            if (
                not isinstance(the_key, CredentialByteArray)
                or len(the_key) != self.key_byte_size
            ):
                raise ValueError(
                    f"The key must be the specified length of {self.key_bit_size} bits"
                )
        self.the_key = the_key

    def prepare_for_new_password(self):
        if self.salt:
            self.salt.zero_array()
        self.salt = CredentialByteArray(os.urandom(self.key_byte_size))
        if self.iv:
            self.iv.zero_array()
        self.iv = CredentialByteArray(os.urandom(AES_CBC_Base.BLOCK_SIZE))

    def encrypt_key(self):
        self.password_auth_hash = self._derive_key_encryption_key()
        enc = AesCbcPaddingEncryptor(self.key_encryption_key, self.iv)
        encrypted_key_bytes = enc.update(input_bytes=self.the_key) + enc.finalize()
        self.encrypted_key = CredentialByteArray(encrypted_key_bytes)

    def decrypt_key(self):
        if self.encrypted_key is None:
            raise ValueError("No encrypted key to decrypt.")
        derived_password_auth_hash = self._derive_key_encryption_key()
        if self.password_auth_hash != derived_password_auth_hash:
            raise PasswordAuthenticationFailure(
                "The password and/or stored secrets do not match."
            )
        dec = AesCbcPaddingDecryptor(key=self.key_encryption_key, IV=self.iv)
        decrypted_key_bytes = (
            dec.update(input_bytes=self.encrypted_key) + dec.finalize()
        )
        self.the_key = CredentialByteArray(decrypted_key_bytes)

    def _derive_key_encryption_key(self):
        if self.password is None:
            raise ValueError("A password is required to derive the key encryption key.")
        if self.salt is None:
            raise ValueError("A salt is required to derive the key encryption key.")
        if self.work_factor is None:
            raise ValueError("A work factor is needed for PBDKF2.")
        self.pbkdf2 = CredentialByteArray(
            pbkdf2_hmac(
                hash_name="sha512",
                password=self.password,
                salt=self.salt,
                iterations=self.work_factor,
                dklen=int(512 / 8),
            )
        )
        self.key_encryption_key = self.pbkdf2.get_portion(0, self.key_byte_size)
        password_auth_hash = self.pbkdf2.get_portion(
            self.key_byte_size, self.key_byte_size * 2
        )
        return password_auth_hash

    def get_as_bytes(
        self,
        include_work_factor: bool = False,
        include_salt: bool = False,
        include_password_auth_hash: bool = False,
        include_IV: bool = False,
        include_key_encryption_key: bool = False,
        include_key: bool = False,
        include_encrypted_key: bool = False,
    ) -> CredentialByteArray:

        #
        # In the following, we ultiamtely get everything into a textual/hex format
        # where fields are separated by the equal sign (=) and commas (,). After
        # TODO: Loop back around to encode with struct pack/unpack.
        #

        r = CredentialByteArray()

        if include_work_factor:
            if len(r) > 0:
                r += b","
            r += b"W="
            r += (
                int(PBKDF2_WORK_FACTOR)
                .to_bytes(length=4, byteorder="little")
                .hex()
                .encode()
            )

        if include_salt:
            if len(r) > 0:
                r += b","
            r += b"S="
            r += self.salt.hex().encode()

        if include_password_auth_hash:
            if len(r) > 0:
                r += b","
            r += b"H="
            r += self.password_auth_hash.hex().encode()

        if include_key_encryption_key:
            if len(r) > 0:
                r += b","
            r += b"E="
            r += self.key_encryption_key.hex().encode()

        if include_IV:
            if len(r) > 0:
                r += b","
            r += b"I="
            r += self.iv.hex().encode()

        if include_encrypted_key:
            if len(r) > 0:
                r += b","
            r += b"C="
            r += self.encrypted_key.hex().encode()

        if include_key:
            if len(r) > 0:
                r += b","
            r += b"K="
            r += self.the_key.hex().encode()

        if len(r) == 0:
            raise ValueError(f"Requesting no secrets.")
        return r

    def set_from_bytes(self, cred_bytes: CredentialByteArray):
        cred_dict = dict(
            kv_pair.split("=") for kv_pair in cred_bytes.decode().split(",")
        )
        for k, v in cred_dict.items():
            c: CredentialByteArray = CredentialByteArray(bytearray.fromhex(v))
            if k == "W":
                self.work_factor = int.from_bytes(bytes=c, byteorder="little")
                if self.work_factor != PBKDF2_WORK_FACTOR:
                    # Until handling for this is in place.
                    raise CredentialInvalid(
                        f"The PBDKF2 work factor {self.work_factor} is invalid. Only {PBKDF2_WORK_FACTOR} is supported."
                    )
            elif k == "S":
                self.salt = c
            elif k == "H":
                self.password_auth_hash = c
            elif k == "E":
                self.key_encryption_key = c
            elif k == "I":
                self.iv = c
            elif k == "C":
                self.encrypted_key = c
            elif k == "K":
                self.the_key = c
            else:
                raise ValueError(f"Cannot interpret value code '{k}'.")

    @staticmethod
    def create_credential_from_bytes(cred_bytes: CredentialByteArray):
        c = Credential()
        c.set_from_bytes(cred_bytes=cred_bytes)
        return c


def prompt_for_password(
    prompt, prompt_again=None, hidden: bool = True
) -> CredentialByteArray:
    pw_input_func = pwinput.pwinput if hidden else input
    while True:
        password_attempt1 = CredentialByteArray(pw_input_func(prompt).encode())
        if len(password_attempt1) == 0:
            print("Blank passwords are not allowed, try again.")
            continue
        if not prompt_again:
            return password_attempt1
        password_attempt2 = CredentialByteArray(pw_input_func(prompt_again).encode())
        if password_attempt1 == password_attempt2:
            del password_attempt2
            return password_attempt1
        else:
            print("The passwords did not match, try again.")


def get_base64_password_from_keyring(
    service_name: str,
    username: str,
) -> tuple[str, str, CredentialByteArray]:
    raw_password = keyring.get_password(service_name=service_name, username=username)
    if raw_password is None:
        raise CredentialNotFoundError(
            f"The credential was not found in the keyring: service_name={service_name} username={username}"
        )
    password_parts = raw_password.split(":", maxsplit=1)
    if len(password_parts) != 2:
        raise CredentialInvalid(
            f"The stored credential is not in the expected two-part format."
        )
    password_type_code = password_parts[0]
    cba_password_base64 = CredentialByteArray(password_parts[1].encode("utf-8"))

    if len(password_type_code) == 0:
        raise CredentialInvalid(
            f"The stored credential is not in the expected format. "
            f"The password type code is not present. "
            f"service_name={service_name} username={username}"
        )

    if len(cba_password_base64) == 0:
        raise CredentialInvalid(
            f"The stored credential is not in the expected format. "
            f"The password is not present. "
            f"service_name={service_name} username={username}"
        )

    if not is_valid_base64_string(cba_password_base64):
        raise InvalidBase64StringError(
            f"Expected stored credential to be in base64 format."
        )

    if password_type_code not in PASSWORD_TYPE_CHAR_TO_TYPE:
        raise CredentialInvalid(
            f"The stored credential is not in the expected format. "
            f"The password type code '{password_type_code}' is unknown. "
            f"service_name={service_name} username={username}"
        )

    password_type = PASSWORD_TYPE_CHAR_TO_TYPE[password_type_code]

    return (password_type_code, password_type, cba_password_base64)


def unlock_credential(credential: Credential):
    """Unlock a Credential instance which is to make its private key/secret
    available in the clear. If the private key is password-protected, ask
    for a password. The result of calling this function is an unlocked
    credential or an exception.

        Exceptions:
            PasswordAuthenticationFailure: Too many attempts with incorrect password.
            CredentialSecretDerivationError: Something unexpected indicates the key
            is not available despite having been decrypted (likely a program bug or
            something esoteric such as system corruption).
    """
    if not credential.is_private_key_possible:
        raise CredentialInvalid(
            f"The private key is not available. "
            f"The credential is invalid or corrupt."
        )
    if not credential.is_private_key_ready:
        if not credential.is_password_required:
            credential.decrypt_key()
        else:
            attempts = 5
            while True:
                password = prompt_for_password(
                    prompt="Enter a password for this backup:"
                )
                credential.set(password=password)
                try:
                    attempts -= 1
                    credential.decrypt_key()
                    break
                except PasswordAuthenticationFailure:
                    if attempts > 0:
                        print(f"The password appears to be invalid, try again.")
                    else:
                        print(f"Still incorrect.")
                        raise
        if not credential.is_private_key_ready:
            raise CredentialSecretDerivationError(
                "Unexpected failure, canonot access the credential."
            )


def get_password_from_keyring(
    service_name: str,
    username: str,
    keep_secret_base64_encoded: bool,
) -> tuple[str, str, CredentialByteArray]:
    (
        password_type_code,
        password_type,
        cba_password_base64,
    ) = get_base64_password_from_keyring(
        service_name=service_name,
        username=username,
    )

    if keep_secret_base64_encoded:
        return (password_type_code, password_type, cba_password_base64)

    cba_password = CredentialByteArray(base64.b64decode(cba_password_base64))

    #
    # Caller wants direct secret in the clear.
    # Secret has been base64 decoded.
    # Enryption secrets are persisted Credential information.
    # To give caller direct secret, perform the following...
    #

    if (
        username == CONFIG_KEYRING_USERNAME_BACKUP_ENCRYPTION
        and password_type == CONFIG_PASSWORD_TYPE_ACTUAL
    ):
        credential = Credential.create_credential_from_bytes(cred_bytes=cba_password)
        unlock_credential(credential=credential)
        cba_password = credential.the_key
        del credential

    return (password_type_code, password_type, cba_password)


def get_enc_credential_from_keyring(
    service_name: str, username: str, unlock: bool = False
) -> Credential:

    if username != CONFIG_KEYRING_USERNAME_BACKUP_ENCRYPTION:
        raise CredentialRequestInvalidError(
            f"A request for a Credential instance is only valid for the backup encryption secret."
        )

    (
        password_type_code,
        password_type,
        cba_password_base64,
    ) = get_base64_password_from_keyring(
        service_name=service_name,
        username=username,
    )

    if password_type != CONFIG_PASSWORD_TYPE_ACTUAL:
        raise CredentialTypeNotFoundError(
            f"Expected credential type={CONFIG_PASSWORD_TYPE_ACTUAL} but got {password_type_code}. "
            f"Cannot derive Credential instance."
        )

    cba_password = CredentialByteArray(base64.b64decode(cba_password_base64))
    credential = Credential.create_credential_from_bytes(cred_bytes=cba_password)
    if not credential.is_private_key_possible:
        raise CredentialInvalid(
            f"The private key is not available. The credential is invalid or corrupt."
        )

    if unlock:
        unlock_credential(credential=credential)

    return credential


def set_password_to_keyring(
    service_name: str,
    username: str,
    password_type: str,
    password_bytes: CredentialByteArray,
    password_is_base64: bool = False,
) -> None:
    if not service_name:
        raise CredentialInvalid(
            f"Cannot store credential. The supplied service name must be a non-empty string."
        )

    if not username:
        raise CredentialInvalid(
            f"Cannot store credential. The supplied user name must be a non-empty string."
        )

    if not password_type:
        raise CredentialInvalid(
            f"Cannot store credential. The supplied password type must be a non-empty string."
        )

    password_type_code = password_type[0]
    if password_type_code not in PASSWORD_TYPE_CHAR_TO_TYPE:
        raise CredentialInvalid(
            f"Cannot store credential. The supplied password type '{password_type}' is invalid."
        )

    if not password_bytes:
        raise CredentialInvalid(
            f"Cannot store credential. The password was not supplied."
        )

    if not password_is_base64:
        password_bytes = CredentialByteArray(base64.b64encode(password_bytes))

    if not is_valid_base64_string(str_to_check=password_bytes):
        raise InvalidBase64StringError(
            f"Expected password_bytes to be a base64-encoded string."
        )

    keyring.set_password(
        service_name=service_name,
        username=username,
        password=f"{password_type_code}:{str(password_bytes, 'utf-8')}",
    )



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging

from ..common.exception import *
from ..common.constants import *
from ..common.file_info import SneakyCorruptionPotential
from .database import FileInformationDatabaseCollection, extract_location_info


def handle_update_digests(args):
    logging.debug("handle_update_digests func")
    if args.locations is None:
        raise ValueError("You must specify one or more locations.")
    logging.debug(f"locations={args.locations}")
    location_persist_type_list = extract_location_info(
        arguments=args.locations,
        min_required=1,
        max_allowed=None,
    )
    total_files = 0
    total_files_created = 0
    total_files_updated = 0
    total_files_no_update_required = 0
    total_files_skipped = 0
    sneaky_corruption_potentials: list[SneakyCorruptionPotential] = []
    try:
        for loc_persist_type in location_persist_type_list:
            location = loc_persist_type[0]
            persist_types = loc_persist_type[1]
            if not os.path.exists(location):
                raise FileNotFoundError(f"Directory does not exist: {location}")
            logging.info(
                f"-------------------------------------------------------------------------"
            )
            logging.info(f"Updating files in {location}...")
            fi_dbc = FileInformationDatabaseCollection(
                source_path=location, persist_types=persist_types
            )
            updater = fi_dbc.update(
                change_detection_type=args.change_detection_type,
                update_stale=True,
                whatif=False,
            )
            logging.info(f"{'Location ':.<45} {location}")
            logging.info(f"    {'Location total files ':.<45} {updater.total_files}")
            logging.info(
                f"    {'Location files info created ':.<45} {updater.total_file_info_created}"
            )
            logging.info(
                f"    {'Location files info updated ':.<45} {updater.total_file_info_updated}"
            )
            logging.info(
                f"    {'Location files no update required ':.<45} {updater.total_files_no_update_required}"
            )
            logging.info(
                f"    {'Location files info stale/error, skipped':.<45} {updater.total_files_stale_info_skipped + updater.total_files_without_primary_digest}"
            )
            if len(updater.sneaky_corruption_potentials) > 0:
                logging.info(
                    f"{'Total potential sneaky corruption':.<45} {len(updater.sneaky_corruption_potentials)} (see details above)"
                )
            total_files += updater.total_files
            total_files_created += updater.total_file_info_created
            total_files_updated += updater.total_file_info_updated
            total_files_no_update_required += updater.total_files_no_update_required
            total_files_skipped += (
                updater.total_files_stale_info_skipped
                + updater.total_files_without_primary_digest
            )
            sneaky_corruption_potentials.extend(updater.sneaky_corruption_potentials)
    except Exception as ex:
        logging.info("")
        logging.error(
            f"ERROR: The update-digest command failed: {os.linesep}{exc_to_string_with_newlines(ex)}"
        )
        logging.info("")
        logging.error(
            "Given the exception above, the following summary may be incomplete:"
        )
        logging.info("")
    finally:
        logging.info(f"=" * 65)
        if len(sneaky_corruption_potentials) > 0:
            logging.info(f"=" * 65)
            logging.info("Potential sneaky corruption all locations processed:")
            for scp in sneaky_corruption_potentials:
                logging.info(f"    path={scp.file_info.path}")
                logging.info(f"    old_size={scp.old_size_in_bytes}")
                logging.info(f"    cur_size={scp.cur_size_in_bytes}")
                logging.info(f"    old_time={scp.old_modified_time}")
                logging.info(f"    cur_time={scp.cur_modified_time}")
                logging.info(f"    old_digest={scp.old_digest}")
                logging.info(f"    cur_digest={scp.cur_digest}")
                logging.info(f"-" * 65)
            logging.info(
                f"    Total potential sneaky corruption Location A: {len(sneaky_corruption_potentials)}"
            )
            logging.info(f"=" * 65)

        logging.info(f"Total all locations processed:")
        logging.info(f"    {'Total files ':.<45} {total_files}")
        logging.info(f"    {'Total Files info created ':.<45} {total_files_created}")
        logging.info(f"    {'Total files info updated ':.<45} {total_files_updated}")
        logging.info(
            f"    {'Total files no update required ':.<45} {total_files_no_update_required}"
        )
        logging.info(
            f"    {'Total files info stale/error, skipped':.<45} {total_files_skipped}"
        )
        logging.info(
            f"    {'Total potential sneaky corruption':.<45} {len(sneaky_corruption_potentials)} (see details above)"
        )



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""ATBU constants.
"""
from typing import Literal


ATBU_MAJOR_VERSION = 0
ATBU_MINOR_VERSION = 1

ATBU_VERSION_STRING = f"{ATBU_MAJOR_VERSION}.{ATBU_MINOR_VERSION:02}"

AT_PREFIX = "at"
AT_PREFIX_U = "AT"
ATBU_ACRONYM = f"{AT_PREFIX}bu"
ATBU_ACRONUM_U = f"{AT_PREFIX_U}BU"

ATBU_PROGRAM_NAME = ATBU_ACRONYM

ATBU_PERSISTENT_INFO_EXTENSION = f".{ATBU_ACRONYM}"
ATBU_PERSISTENT_INFO_DB_EXTENSION = f".{ATBU_ACRONYM}db"
ATBU_SKIP_EXTENSIONS = [
    ATBU_PERSISTENT_INFO_EXTENSION,
    ATBU_PERSISTENT_INFO_DB_EXTENSION,
]
ATBU_DEFAULT_PERSISTENT_DB_FILENAME = (
    f"c4198ead-0b50-4f0e-b52b-685b64e7b9f0{ATBU_PERSISTENT_INFO_DB_EXTENSION}"
)
ATBU_PERSISTENT_INFO_VERSION = 2
ATBU_PERSISTENT_INFO_VERSION_STRING = f"{ATBU_PERSISTENT_INFO_VERSION}"
ATBU_PERSIST_TYPE_PER_DIR = "per-dir"
ATBU_PERSIST_TYPE_PER_FILE = "per-file"
ATBU_PERSIST_TYPE_PER_BOTH = "per-both"
ATBU_PERSIST_TYPES = [ATBU_PERSIST_TYPE_PER_DIR, ATBU_PERSIST_TYPE_PER_FILE]
ATBU_PERSIST_TYPE_HINT = Literal["per-dir", "per-file"]
ATBU_PERSIST_TYPE_CHAR_TO_STR = {
    "d": ATBU_PERSIST_TYPE_PER_DIR,
    "f": ATBU_PERSIST_TYPE_PER_FILE,
    "b": ATBU_PERSIST_TYPE_PER_BOTH,
}

ATBU_FILE_BACKUP_EXTENSION = f".{AT_PREFIX}bak"
ATBU_FILE_BACKUP_EXTENSION_ENCRYPTED = f".{AT_PREFIX}bake"

ATBU_BACKUP_TYPE_FULL = "full"
ATBU_BACKUP_TYPE_INCREMENTAL = "incremental"
ATBU_BACKUP_TYPE_INCREMENTAL_PLUS_SHORT_CMD_OPT = "ip"
ATBU_BACKUP_TYPE_INCREMENTAL_PLUS = "incremental-plus"
ATBU_BACKUP_TYPE_ALL = [
    ATBU_BACKUP_TYPE_FULL,
    ATBU_BACKUP_TYPE_INCREMENTAL,
    ATBU_BACKUP_TYPE_INCREMENTAL_PLUS,
]

ATBU_BACKUP_DEDUPLICATION_TYPE_DIGEST = "digest"
ATBU_BACKUP_DEDUPLICATION_TYPE_DIGEST_EXT = "digest-ext"

BACKUP_COMPRESSION_NONE = "none"
BACKUP_COMPRESSION_NORMAL = "normal"
BACKUP_COMPRESSION_CHOICES = [BACKUP_COMPRESSION_NONE, BACKUP_COMPRESSION_NORMAL]
BACKUP_COMPRESSION_DEFAULT = BACKUP_COMPRESSION_NORMAL
BACKUP_COMPRESSION_TYPE = "gzip"

CONFIG_SECTION_COMPRESSION = "compression"
CONFIG_VALUE_NAME_COMPRESSION_LEVEL = "level"
CONFIG_VALUE_NAME_NO_COMPRESS_PATTERN = "no_compress_pattern"
ATBU_BACKUP_DEFAULT_NO_COMPRESS_RE_PAT = (
    r"("
    r".*\.jpg$|.*\.jpeg$|.*\.mp4$|.*\.mov$|.*\.mpg|.*\.mpeg$|.*\.mp3$|"
    r".*\.zip$|.*\.gz$|.*\.7z$|.*\.bz2$"
    r")"
)
CONFIG_VALUE_NAME_COMPRESS_MIN_FILE_SIZE = "min_size"
ATBU_BACKUP_DEFAULT_COMPRESS_MIN_FILE_SIZE = 150
CONFIG_VALUE_NAME_MIN_COMPRESS_RATIO = "min_ratio"
ATBU_BACKUP_DEFAULT_MIN_COMPRESS_RATIO = 0.9
CONFIG_VALUE_NAME_MAX_FTYPE_ATTEMPTS = "max_file_type_attempts"
ATBU_BACKUP_DEFAULT_MAX_FTYPE_ATTEMPTS = 3

ATBU_BACKUP_COMPRESSION_DEFAULTS = {
    CONFIG_VALUE_NAME_COMPRESSION_LEVEL: BACKUP_COMPRESSION_NORMAL,
    CONFIG_VALUE_NAME_NO_COMPRESS_PATTERN: ATBU_BACKUP_DEFAULT_NO_COMPRESS_RE_PAT,
    CONFIG_VALUE_NAME_COMPRESS_MIN_FILE_SIZE: ATBU_BACKUP_DEFAULT_COMPRESS_MIN_FILE_SIZE,
    CONFIG_VALUE_NAME_MIN_COMPRESS_RATIO: ATBU_BACKUP_DEFAULT_MIN_COMPRESS_RATIO,
    CONFIG_VALUE_NAME_MAX_FTYPE_ATTEMPTS: ATBU_BACKUP_DEFAULT_MAX_FTYPE_ATTEMPTS,
}

ATBU_CONFIG_NAME = f"{ATBU_ACRONUM_U} Configuration"

ATBU_CONFIG_FILE_MAJOR_VERSION = 0
ATBU_CONFIG_FILE_MINOR_VERSION = 1
ATBU_CONFIG_FILE_VERSION_STRING = (
    f"{ATBU_CONFIG_FILE_MAJOR_VERSION}.{ATBU_CONFIG_FILE_MINOR_VERSION:02}"
)

ATBU_DEFAULT_CONFIG_DIR_NAME = f".{ATBU_ACRONYM}"
ATBU_DEFAULT_CONFIG_FILE_NAME = f"{ATBU_ACRONYM}-config.json"
ATBU_DEFAULT_BACKUP_INFO_SUBDIR = f"{ATBU_ACRONYM}-backup-info"

CONFIG_SECTION_GENERAL = "general"
CONFIG_VALUE_NAME_BACKUP_INFO_DIR = "backup-info-dir"
CONFIG_VALUE_NAME_CONFIG_NAME = "name"
CONFIG_VALUE_NAME_VERSION = "version"

CONFIG_SECTION_STORAGE_DEFINITIONS = "storage-definitions"
CONFIG_SECTION_STORAGE_DEFINITION_SPECIFIER_PREFIX = "storage-def"
CONFIG_VALUE_NAME_INTERFACE_TYPE = "interface"
CONFIG_VALUE_NAME_PROVIDER = "provider"
CONFIG_VALUE_NAME_CONTAINER = "container"
CONFIG_SECTION_DRIVER = "driver"
CONFIG_VALUE_NAME_DRIVER_STORAGE_KEY = "key"
CONFIG_VALUE_NAME_DRIVER_STORAGE_SECRET = "secret"
CONFIG_VALUE_NAME_DRIVER_STORAGE_PROJECT = "project"
CONFIG_SECTION_KEYRING_MAPPING = "keyring-mapping"
CONFIG_SECTION_KEYRING_MAPPING_ENCRYPTION_KEY = "encryption-key"
CONFIG_VALUE_NAME_KEYRING_SERVICE = "service"
CONFIG_VALUE_NAME_KEYRING_USERNAME = "username"
CONFIG_SECTION_ENCRYPTION = "encryption"
CONFIG_VALUE_NAME_ENCRYPTION_KEY = "key"
CONFIG_VALUE_NAME_STORAGE_PERSISTED_IV = "storage-persisted-IV"
CONFIG_STORAGE_PERSISTED_IV_DEFAULT_VALUE = True
CONFIG_VALUE_NAME_STORAGE_PERSISTED_BACKUP_INFO = "storage-persisted-backup-info"
CONFIG_STORAGE_PERSISTED_BACKUP_INFO_DEFAULT_VALUE = True
CONFIG_KEY_VALUE_KEYRING_INDIRECTION = "keyring"
CONFIG_KEYRING_USERNAME_BACKUP_ENCRYPTION = f"{ATBU_ACRONUM_U}-backup-enc-key"
CONFIG_KEYRING_USERNAME_STORAGE_PASSWORD = f"{ATBU_ACRONUM_U}-storage-password"

CONFIG_INTERFACE_TYPE_LIBCLOUD = "libcloud"
CONFIG_INTERFACE_TYPE_GOOGLE = "google"
CONFIG_INTERFACE_TYPE_FILESYSTEM = "filesystem"
CONFIG_INTERFACE_HINT = Literal["libcloud", "google", "filesystem"]

CONFIG_PASSWORD_TYPE = "password_type"
# CONFIG_CREDENTIAL_TYPE will be one of...
CONFIG_PASSWORD_TYPE_ACTUAL = "actual"
CONFIG_PASSWORD_TYPE_FILENAME = "filename"
CONFIG_PASSWORD_TYPE_ENVVAR = "envvar"

CRED_OPERATION_SET_PASSWORD = "set-password"
CRED_OPERATION_SET_PASSWORD_TO_FILENAME = "set-password-filename"
CRED_OPERATION_SET_PASSWORD_TO_ENVVAR = "set-password-envvar"
CRED_OPERATION_SET_PASSWORD_TO_PRIVATE_KEY = "set-password-private-key"
CRED_OPERATION_HINT = Literal[
    "set-password",
    "set-password-filename",
    "set-password-envvar",
    "set-password-private-key",
]

CRED_KEY_TYPE_STORAGE = "storage-secret"
CRED_KEY_TYPE_ENCRYPTION = "encryption-key"
CRED_KEY_TYPE_HINT = Literal["storage-secret", "encryption-key"]

PASSWORD_TYPE_CHAR_TO_TYPE = {
    i[0]: i
    for i in [
        CONFIG_PASSWORD_TYPE_ACTUAL,
        CONFIG_PASSWORD_TYPE_FILENAME,
        CONFIG_PASSWORD_TYPE_ENVVAR,
    ]
}

ALLOWED_KEY_BIT_LENGTHS = [128, 192, 256]
DEFAULT_KEY_BIT_LENGTH = 256
# On 2021 16-core / 32-thread die, 1 million takes about .5 seconds.
# Obviously a little longer on 2017 mobile device but still acceptable
# given the relative infrequent need for password setup.
PBKDF2_WORK_FACTOR = 1000000

# BASE64_PREFIX_ORIGINAL_IS_BYTES = 0x0B
# BASE64_PREFIX_ORIGINAL_IS_STR = 0x05



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""An instance of HasherDefinitions is configured with one or more
hashing algorithms, where it can create a Hasher instance that can
produce a hash for each algorithm simultaneously. Currently, only a
single algorithm of SHA256 is used but there is room to expand out
as needed (there just was not time during initial creation, so desire
was to focus on SHA256 as the so-called "primary" hash algo).
"""
import hashlib

from .singleton import Singleton
from .exception import SingletonAlreadyCreated

DEFAULT_HASH_ALGORITHM = "sha256"


class Hasher:
    def __init__(self, hashers):
        self.hashers = hashers

    def __iter__(self):
        return self.hashers

    def get_primary_hashing_algo_name(self):
        return self.hashers[0][0]

    def update_all(self, data):
        for h in self.hashers:
            h[1].update(data)

    def get_digests(self):
        r = {}
        for h in self.hashers:
            r[h[0]] = h[1].digest()
        return r

    def get_hexdigests(self):
        r = {}
        for h in self.hashers:
            r[h[0]] = h[1].hexdigest()
        return r

    def get_primary_hexdigest(self) -> str:
        return self.get_hexdigests()[self.get_primary_hashing_algo_name()]


class HasherDefinitions:
    def __init__(self, algos: list[str] = None):
        if algos is None:
            algos = [DEFAULT_HASH_ALGORITHM]
        self.hash_classes = []
        for a in algos:
            c = getattr(hashlib, a)
            if c is None:
                raise NameError(f"Algorithm unknown: {a}")
            self.hash_classes.append((a, c))

    def create_hasher(self) -> Hasher:
        r = []
        for c in self.hash_classes:
            r.append((c[0].lower(), c[1]()))
        return Hasher(r)

    def get_primary_hashing_algo_name(self):
        return self.hash_classes[0][0].lower()


class GlobalHasherDefinitions(HasherDefinitions, metaclass=Singleton):
    _init_algos: list[str] = None

    def __init__(self, algos: list[str] = None):
        if self._init_algos:
            if algos:
                raise SingletonAlreadyCreated(
                    "Access to singleton after initial creation must have no arguments."
                )
            return  # do not call base class __init__ after first-time init.
        if not algos:
            algos = [DEFAULT_HASH_ALGORITHM]
        if not isinstance(algos, list):
            raise ValueError(
                f"GlobalHasherDefinitions: The algos argument must be a list of algorithms."
            )
        self._init_algos = algos
        super().__init__(algos)
